import os
import collections
from glob import glob
from .data import Data, WARP
from .helpers import itemize, expandDir, collectFormats, cleanName
from .timestamp import Timestamp
from .prepare import (levels, order, rank, levUp, levDown, boundary, sections)
from .api import (
    Api,
    NodeFeature,
    EdgeFeature,
    OtypeFeature,
    OslotsFeature,
    Computed,
    addSortKey,
    addOtype,
    addLocality,
    addText,
    addSearch,
)
from .mql import MQL, tfFromMql

NAME = 'Text-Fabric'
VERSION = '3.4.12'
APIREF = 'https://github.com/Dans-labs/text-fabric/wiki/Api'
TUTORIAL = (
    'https://github.com/Dans-labs/text-fabric'
    '/blob/master/docs/tutorial.ipynb'
)
DATA = 'https://github.com/Dans-labs/text-fabric-data'

LOCATIONS = [
    '~/Downloads/text-fabric-data',
    '~/text-fabric-data',
    '~/github/text-fabric-data',
    '~/Dropbox/text-fabric-data',
    '/mnt/shared/text-fabric-data',
]

DATASETS = [
    'hebrew/etcbc4c',
]

MODULES = [
    'core',
    'phono',
]

PRECOMPUTE = (
    (False, '__levels__', levels, WARP),
    (False, '__order__', order, WARP[0:2] + ('__levels__', )),
    (False, '__rank__', rank, (WARP[0], '__order__')),
    (False, '__levUp__', levUp, WARP[0:2] + ('__rank__', )),
    (False, '__levDown__', levDown, (WARP[0], '__levUp__', '__rank__')),
    (False, '__boundary__', boundary, WARP[0:2] + ('__rank__', )),
    (True, '__sections__', sections, WARP + ('__levUp__', '__levels__')),
)


class Fabric(object):
    def __init__(self, locations=None, modules=None, silent=False):
        self.silent = silent
        self.tm = Timestamp()
        self.banner = f'This is {NAME} {VERSION}'
        self.version = VERSION
        if not silent:
            self.tm.info(
                '''{}
Api reference : {}
Tutorial      : {}
Example data  : {}
'''.format(
                    self.banner,
                    APIREF,
                    TUTORIAL,
                    DATA,
                ),
                tm=False
            )
        self.good = True

        if modules is None:
            modules = ['']
        if type(modules) is str:
            modules = [x.strip() for x in itemize(modules, '\n')]
        self.modules = modules

        if locations is None:
            locations = LOCATIONS
        if type(locations) is str:
            locations = [x.strip() for x in itemize(locations, '\n')]
        self.homeDir = os.path.expanduser('~').replace('\\', '/')
        self.curDir = os.getcwd().replace('\\', '/')
        (self.parentDir, x) = os.path.split(self.curDir)
        self.locations = []
        for loc in locations:
            self.locations.append(
                expandDir(
                    loc,
                    dict(
                        cur=self.curDir,
                        up=self.parentDir,
                        home=self.homeDir,
                    )
                )
            )

        self.locationRep = '\n\t'.join(
            '\n\t'.join('{}/{}'.format(l, f) for f in self.modules)
            for l in self.locations
        )
        self.featuresRequested = []
        self._makeIndex()

    def load(self, features, add=False, silent=False):
        self.tm.indent(level=0, reset=True)
        if not silent:
            self.tm.info('loading features ...')
        self.sectionsOK = True
        self.good = True
        if self.good:
            featuresRequested = itemize(
                features
            ) if type(features) is str else sorted(features)
            if add:
                self.featuresRequested += featuresRequested
            else:
                self.featuresRequested = featuresRequested
            for fName in list(WARP):
                self._loadFeature(
                    fName, optional=fName == WARP[2], silent=silent
                )
        if self.good:
            self._cformats = {}
            self._formatFeats = []
            if WARP[2] in self.features:
                otextMeta = self.features[WARP[2]].metaData
                for otextMod in self.features:
                    if otextMod.startswith(WARP[2] + '@'):
                        self._loadFeature(otextMod, silent=silent)
                        otextMeta.update(self.features[otextMod].metaData)
                sectionFeats = itemize(
                    otextMeta.get('sectionFeatures', ''), ','
                )
                sectionTypes = itemize(otextMeta.get('sectionTypes', ''), ',')
                if len(sectionTypes) != 3 or len(sectionFeats) != 3:
                    if not silent:
                        self.tm.info(
                            'Not enough info for sections in'
                            ' {}, section functionality will not work'.format(
                                WARP[2]
                            )
                        )
                    self.sectionsOK = False
                else:
                    for (i, fName) in enumerate(sectionFeats):
                        self._loadFeature(fName, silent=silent)
                if self.good:
                    (cformats, formatFeats) = collectFormats(otextMeta)
                    for fName in formatFeats:
                        self._loadFeature(fName, silent=silent)
                    self._cformats = cformats
                    self._formatFeats = formatFeats
            else:
                self.sectionsOK = False

        if self.good:
            self._precompute()
        if self.good:
            for fName in self.featuresRequested:
                self._loadFeature(fName, silent=silent)
        if not self.good:
            self.tm.indent(level=0)
            self.tm.error('Not all features could be loaded/computed')
            self.tm.cache()
            return None
        if add:
            self._updateApi(silent)
        else:
            return self._makeApi(silent)

    def explore(self, silent=True, show=True):
        nodes = set()
        edges = set()
        configs = set()
        computeds = set()
        for (fName, fObj) in self.features.items():
            fObj.load(metaOnly=True, silent=True)
            dest = None
            if fObj.method:
                dest = computeds
            elif fObj.isConfig:
                dest = configs
            elif fObj.isEdge:
                dest = edges
            else:
                dest = nodes
            dest.add(fName)
        if not silent:
            self.tm.info((
                'Feature overview: {} for nodes;'
                ' {} for edges; {} configs; {} computed'
            ).format(
                len(nodes),
                len(edges),
                len(configs),
                len(computeds),
            ))
        self.featureSets = dict(
            nodes=nodes, edges=edges, configs=configs, computeds=computeds
        )
        if show:
            return dict(
                (kind, tuple(sorted(kindSet)))
                for (kind, kindSet
                     ) in sorted(self.featureSets.items(), key=lambda x: x[0])
            )

    def clearCache(self):
        for (fName, fObj) in self.features.items():
            fObj.cleanDataBin()

    def save(self, nodeFeatures={}, edgeFeatures={}, metaData={}, module=None):
        self.tm.indent(level=0, reset=True)
        self._getWriteLoc(module=module)
        configFeatures = dict(
            f for f in metaData.items()
            if f[0] != '' and f[0] not in nodeFeatures
            and f[0] not in edgeFeatures
        )
        if not self.silent:
            self.tm.info(
                'Exporting {} node and {} edge and {} config features to {}:'.
                format(
                    len(nodeFeatures),
                    len(edgeFeatures),
                    len(configFeatures),
                    self.writeDir,
                )
            )
        todo = []
        for (fName, data) in sorted(nodeFeatures.items()):
            todo.append((fName, data, False, False))
        for (fName, data) in sorted(edgeFeatures.items()):
            todo.append((fName, data, True, False))
        for (fName, data) in sorted(configFeatures.items()):
            todo.append((fName, data, None, True))
        total = collections.Counter()
        failed = collections.Counter()
        for (fName, data, isEdge, isConfig) in todo:
            edgeValues = False
            fMeta = {}
            fMeta.update(metaData.get('', {}))
            fMeta.update(metaData.get(fName, {}))
            if fMeta.get('edgeValues', False):
                edgeValues = True
            if 'edgeValues' in fMeta:
                del fMeta['edgeValues']
            fObj = Data(
                '{}/{}.tf'.format(self.writeDir, fName),
                self.tm,
                data=data,
                metaData=fMeta,
                isEdge=isEdge,
                isConfig=isConfig,
                edgeValues=edgeValues,
            )
            tag = 'config' if isConfig else 'edge' if isEdge else 'node'
            if fObj.save(nodeRanges=fName == WARP[0], overwrite=True):
                total[tag] += 1
            else:
                failed[tag] += 1
        self.tm.indent(level=0)
        if not self.silent:
            self.tm.info((
                'Exported {} node features and {} edge features'
                ' and {} config features to {}'
            ).format(
                total['node'],
                total['edge'],
                total['config'],
                self.writeDir,
            ))
        if len(failed):
            for (tag, nf) in sorted(failed.items()):
                self.tm.error(
                    'Failed to export {} {} features'.format(nf, tag)
                )

    def exportMQL(self, mqlName, mqlDir):
        self.tm.indent(level=0, reset=True)
        mqlDir = expandDir(
            mqlDir,
            dict(
                cur=self.curDir,
                up=self.parentDir,
                home=self.homeDir,
            )
        )

        mqlNameClean = cleanName(mqlName)
        mql = MQL(mqlDir, mqlNameClean, self.features, self.tm)
        mql.write()

    def importMQL(self, mqlFile, slotType=None, otext=None, meta=None):
        self.tm.indent(level=0, reset=True)
        (good, nodeFeatures, edgeFeatures, metaData) = tfFromMql(
            mqlFile, self.tm, slotType=slotType, otext=otext, meta=meta
        )
        if good:
            self.save(
                nodeFeatures=nodeFeatures,
                edgeFeatures=edgeFeatures,
                metaData=metaData
            )

    def _loadFeature(self, fName, optional=False, silent=False):
        if not self.good:
            return False
        if fName not in self.features:
            if not optional:
                self.tm.error(
                    'Feature "{}" not available in\n{}'.format(
                        fName, self.locationRep
                    )
                )
                self.good = False
        else:
            if not self.features[fName].load(
                silent=silent or (fName not in self.featuresRequested)
            ):
                self.good = False

    def _makeIndex(self):
        self.features = {}
        self.featuresIgnored = {}
        tfFiles = {}
        for loc in self.locations:
            for mod in self.modules:
                files = glob('{}/{}/*.tf'.format(loc, mod))
                for f in files:
                    if not os.path.isfile(f):
                        continue
                    (dirF, fileF) = os.path.split(f)
                    (fName, ext) = os.path.splitext(fileF)
                    tfFiles.setdefault(fName, []).append(f)
        for (fName, featurePaths) in sorted(tfFiles.items()):
            chosenFPath = featurePaths[-1]
            for featurePath in sorted(set(featurePaths[0:-1])):
                if featurePath != chosenFPath:
                    self.featuresIgnored.setdefault(fName,
                                                    []).append(featurePath)
            self.features[fName] = Data(chosenFPath, self.tm)
        self._getWriteLoc()
        if not self.silent:
            self.tm.info(
                '{} features found and {} ignored'.format(
                    len(tfFiles),
                    sum(len(x) for x in self.featuresIgnored.values()),
                ),
                tm=False
            )

        good = True
        for fName in WARP:
            if fName not in self.features:
                if fName == WARP[2]:
                    if not self.silent:
                        self.tm.info((
                            'Warp feature "{}" not found.'
                            ' Working without Text-API\n'
                        ).format(WARP[2]))
                else:
                    if not self.silent:
                        self.tm.error(
                            'Warp feature "{}" not found in\n{}'.format(
                                fName, self.locationRep
                            )
                        )
                    good = False
            elif fName == WARP[2]:
                self._loadFeature(fName, optional=True, silent=True)
        if not good:
            return False
        self.warpDir = self.features[WARP[0]].dirName
        self.precomputeList = []
        for (dep2, fName, method, dependencies) in PRECOMPUTE:
            thisGood = True
            if dep2 and WARP[2] not in self.features:
                continue
            if dep2:
                otextMeta = self.features[WARP[2]].metaData
                sectionFeats = tuple(
                    itemize(otextMeta.get('sectionFeatures', ''), ',')
                )
                dependencies = dependencies + sectionFeats
            for dep in dependencies:
                if dep not in self.features:
                    self.tm.error(
                        'Missing dependency for computed data feature '
                        '"{}": "{}"'.format(fName, dep)
                    )
                    thisGood = False
            if not thisGood:
                good = False
            self.features[fName] = Data(
                '{}/{}.x'.format(self.warpDir, fName),
                self.tm,
                method=method,
                dependencies=[
                    self.features.get(dep, None) for dep in dependencies
                ],
            )
            self.precomputeList.append((fName, dep2))
        self.good = good

    def _getWriteLoc(self, dirName=None, module=None):
        writeLoc = dirName if dirName is not None else '' if len(
            self.locations
        ) == 0 else self.locations[-1]
        writeMod = module if module is not None else '' if len(
            self.modules
        ) == 0 else self.modules[-1]
        self.writeDir = '{}{}'.format(
            writeLoc, writeMod
        ) if writeLoc == '' or writeMod == '' else '{}/{}'.format(
            writeLoc, writeMod
        )

    def _precompute(self):
        good = True
        for (fName, dep2) in self.precomputeList:
            if dep2 and not self.sectionsOK:
                continue
            if not self.features[fName].load(silent=True):
                good = False
                break
        self.good = good

    def _makeApi(self, silent):
        if not self.good:
            return None
        api = Api(self)

        setattr(api.F, WARP[0], OtypeFeature(api, self.features[WARP[0]].data))
        setattr(
            api.E, WARP[1], OslotsFeature(api, self.features[WARP[1]].data)
        )

        sectionFeats = []
        if WARP[2] in self.features:
            otextMeta = self.features[WARP[2]].metaData
            sectionFeats = itemize(otextMeta.get('sectionFeatures', ''), ',')

        for fName in self.features:
            fObj = self.features[fName]
            if fObj.dataLoaded and not fObj.isConfig:
                if fObj.method:
                    feat = fName.strip('_')
                    ap = api.C
                    if fName in [
                        x[0] for x in self.precomputeList
                        if not x[1] or self.sectionsOK
                    ]:
                        setattr(ap, feat, Computed(api, fObj.data))
                    else:
                        fObj.unload()
                        if hasattr(ap, feat):
                            delattr(api.C, feat)
                else:
                    if fName in self.featuresRequested:
                        if fName in WARP:
                            continue
                        elif fObj.isEdge:
                            setattr(
                                api.E, fName,
                                EdgeFeature(api, fObj.data, fObj.edgeValues)
                            )
                        else:
                            setattr(api.F, fName, NodeFeature(api, fObj.data))
                    else:
                        if (
                                fName in WARP or
                                fName in sectionFeats or
                                fName in self._formatFeats
                        ):
                            continue
                        elif fObj.isEdge:
                            if hasattr(api.E, fName):
                                delattr(api.E, fName)
                        else:
                            if hasattr(api.F, fName):
                                delattr(api.F, fName)
                        fObj.unload()
        addSortKey(api)
        addOtype(api)
        addLocality(api)
        addText(api)
        addSearch(api, silent)
        self.tm.indent(level=0)
        if not silent:
            self.tm.info(
                'All features loaded/computed - for details use loadLog()'
            )
        self.api = api
        return api

    def _updateApi(self, silent):
        if not self.good:
            return None
        api = self.api

        sectionFeats = []
        if WARP[2] in self.features:
            otextMeta = self.features[WARP[2]].metaData
            sectionFeats = itemize(otextMeta.get('sectionFeatures', ''), ',')

        for fName in self.features:
            fObj = self.features[fName]
            if fObj.dataLoaded and not fObj.isConfig:
                if not fObj.method:
                    if fName in self.featuresRequested:
                        if fName in WARP:
                            continue
                        elif fObj.isEdge:
                            if not hasattr(api.E, fName):
                                setattr(
                                    api.E, fName,
                                    EdgeFeature(
                                        api, fObj.data, fObj.edgeValues
                                    )
                                )
                        else:
                            if not hasattr(api.F, fName):
                                setattr(
                                    api.F, fName, NodeFeature(api, fObj.data)
                                )
                    else:
                        if (
                                fName in WARP or
                                fName in sectionFeats or
                                fName in self._formatFeats
                        ):
                            continue
                        elif fObj.isEdge:
                            if hasattr(api.E, fName):
                                delattr(api.E, fName)
                        else:
                            if hasattr(api.F, fName):
                                delattr(api.F, fName)
                        fObj.unload()
        self.tm.indent(level=0)
        if not silent:
            self.tm.info(
                'All additional features loaded - for details use loadLog()'
            )



class Locality(object):
    def __init__(self, api):
        self.api = api

    def u(self, n, otype=None):
        if n <= 0:
            return tuple()
        Fotype = self.api.F.otype
        maxNode = Fotype.maxNode
        if n > maxNode:
            return tuple()
        levUp = self.api.C.levUp.data

        if otype is None:
            return levUp[n - 1]
        else:
            return tuple(m for m in levUp[n - 1] if Fotype.v(m) == otype)

    def d(self, n, otype=None):
        Fotype = self.api.F.otype
        maxSlot = Fotype.maxSlot
        if n <= maxSlot:
            return tuple()
        maxNode = Fotype.maxNode
        if n > maxNode:
            return tuple()

        Eoslots = self.api.E.oslots
        Crank = self.api.C.rank.data
        levDown = self.api.C.levDown.data
        slotType = Fotype.slotType
        if otype is None:
            return sorted(
                levDown[n - maxSlot - 1] + Eoslots.s(n),
                key=lambda m: Crank[m - 1],
            )
        elif otype == slotType:
            return sorted(
                Eoslots.s(n),
                key=lambda m: Crank[m - 1],
            )
        else:
            return tuple(
                m for m in levDown[n - maxSlot - 1] if Fotype.v(m) == otype
            )

    def p(self, n, otype=None):
        if n <= 1:
            return tuple()
        Fotype = self.api.F.otype
        maxNode = Fotype.maxNode
        if n > maxNode:
            return tuple()

        maxSlot = Fotype.maxSlot
        Eoslots = self.api.E.oslots.data
        (firstNode, lastNode) = self.api.C.boundary.data

        myPrev = n - 1 if n <= maxSlot else Eoslots[n - maxSlot - 1][0] - 1
        if myPrev <= 0:
            return ()

        result = lastNode[myPrev - 1] + (myPrev, )

        if otype is None:
            return result
        else:
            return tuple(m for m in result if Fotype.v(m) == otype)

    def n(self, n, otype=None):
        if n <= 0:
            return tuple()
        Fotype = self.api.F.otype
        maxNode = Fotype.maxNode
        maxSlot = Fotype.maxSlot
        if n == maxSlot:
            return tuple()
        if n > maxNode:
            return tuple()

        Eoslots = self.api.E.oslots.data
        (firstNode, lastNode) = self.api.C.boundary.data

        myNext = n + 1 if n < maxSlot else Eoslots[n - maxSlot - 1][-1] + 1
        if myNext > maxSlot:
            return ()

        result = (myNext, ) + firstNode[myNext - 1]

        if otype is None:
            return result
        else:
            return tuple(m for m in result if Fotype.v(m) == otype)



import re
import types
import collections
from functools import reduce
from random import randrange
from inspect import signature

from .data import WARP

STRATEGY = '''
    small_choice_first
    spread_1_first
    big_choice_first
'''.strip().split()

escapes = (
    '\\\\',
    '\\ ',
    '\\t',
    '\\n',
    '\\|',
    '\\=',
)

rePat = '^([a-zA-Z0-9-_]+)~(.*)$'
reRe = re.compile(rePat)
reTp = type(reRe)


def esc(x):
    for (i, c) in enumerate(escapes):
        x = x.replace(c, chr(i))
    return x


def unesc(x):
    for (i, c) in enumerate(escapes):
        x = x.replace(chr(i), c[1])
    return x


class Search(object):
    def __init__(self, api, silent):
        self.api = api
        self.good = True
        self.silent = silent
        self._basicRelations()

# API METHODS ###

    def search(self, searchTemplate, limit=None):
        self.silent = True
        self.study(searchTemplate, silent=True)
        return self.fetch(limit=limit)

    def study(self, searchTemplate, strategy=None, silent=False):
        info = self.api.info
        self.silent = silent
        self.api.indent(level=0, reset=True)
        self.good = True

        self._setStrategy(strategy)
        if not self.good:
            return

        if not self.silent:
            info('Checking search template ...')
        self.searchTemplate = searchTemplate

        self._parse()
        self._prepare()
        if not self.good:
            return
        if not self.silent:
            info(
                'Setting up search space for {} objects ...'.format(
                    len(self.qnodes)
                )
            )
        self._spinAtoms()
        if not self.silent:
            info(
                'Constraining search space with {} relations ...'.format(
                    len(self.qedges)
                )
            )
        self._spinEdges()
        if not self.silent:
            info('Setting up retrieval plan ...')
        self._stitch()
        if self.good:
            if not self.silent:
                info(
                    'Ready to deliver results from {} nodes'.format(
                        sum(len(y) for y in self.yarns.values()),
                    )
                )
            if not self.silent:
                info('Iterate over S.fetch() to get the results', tm=False)
            if not self.silent:
                info('See S.showPlan() to interpret the results', tm=False)

    def fetch(self, limit=None):
        if not self.good:
            return []
        if limit is None:
            return self.results()
        results = []
        for r in self.results():
            results.append(r)
            if len(results) == limit:
                break
        return tuple(results)

    def glean(self, r):
        T = self.api.T
        F = self.api.F
        L = self.api.L
        slotType = F.otype.slotType
        lR = len(r)
        if lR == 0:
            return ''
        fields = []
        for (i, n) in enumerate(r):
            otype = F.otype.v(n)
            words = [n] if otype == slotType else L.d(n, otype=slotType)
            if otype == T.sectionTypes[2]:
                field = '{} {}:{}'.format(*T.sectionFromNode(n))
            elif otype == slotType:
                field = T.text(words)
            elif otype in T.sectionTypes[0:2]:
                field = ''
            else:
                field = '{}[{}{}]'.format(
                    otype,
                    T.text(words[0:5]),
                    '...' if len(words) > 5 else '',
                )
            fields.append(field)
        return ' '.join(fields)

    def count(self, progress=None, limit=None):
        info = self.api.info
        indent = self.api.indent
        indent(level=0, reset=True)

        if not self.good:
            self.error(
                'This search has problems. No results to count.', tm=False
            )
            return

        PROGRESS = 100
        LIMIT = 1000

        if progress is None:
            progress = PROGRESS
        if limit is None:
            limit = LIMIT

        info(
            'Counting results per {} up to {} ...'.format(
                progress,
                limit if limit > 0 else ' the end of the results',
            )
        )
        indent(level=1, reset=True)

        i = 0
        j = 0
        for r in self.results(remap=False):
            i += 1
            j += 1
            if j == progress:
                j = 0
                info(i)
            if limit > 0 and i >= limit:
                break

        indent(level=0)
        info('Done: {} results'.format(i))

    def showPlan(self, details=False):
        if not self.good:
            return
        info = self.api.info
        nodeLine = self.nodeLine
        qedges = self.qedges
        (qs, es) = self.stitchPlan
        if details:
            info(
                'Search with {} objects and {} relations'.format(
                    len(qs), len(es)
                ),
                tm=False
            )
            info(
                'Results are instantiations of the following objects:',
                tm=False
            )
            for q in qs:
                self._showNode(q)
            if len(es) != 0:
                info(
                    'Instantiations are computed'
                    ' along the following relations:',
                    tm=False
                )
                (firstE, firstDir) = es[0]
                (f, rela, t) = qedges[firstE]
                if firstDir == -1:
                    (f, t) = (t, f)
                self._showNode(f, pos2=True)
                for e in es:
                    self._showEdge(*e)
        info(
            'The results are connected to the'
            ' original search template as follows:'
        )

        resultNode = {}
        for q in qs:
            resultNode[nodeLine[q]] = q
        for (i, line) in enumerate(self.searchLines):
            rNode = resultNode.get(i, '')
            info(
                '{:>2} {:<1}{:<2} {}'.format(
                    i,
                    'R' if rNode != '' else '',
                    rNode,
                    line,
                ),
                tm=False
            )

# LOW-LEVEL NODE RELATIONS SEMANTICS ###

    def _basicRelations(self):
        C = self.api.C
        F = self.api.F
        E = self.api.E
        Crank = C.rank.data
        ClevDown = C.levDown.data
        ClevUp = C.levUp.data
        (CfirstSlots, ClastSlots) = C.boundary.data
        Eoslots = E.oslots.data
        slotType = F.otype.slotType
        maxSlot = F.otype.maxSlot

        def equalR(fTp, tTp):
            return lambda n: (n, )

        def spinEqual(fTp, tTp):
            def doyarns(yF, yT):
                x = set(yF) & set(yT)
                return (x, x)

            return doyarns

        def unequalR(fTp, tTp):
            return lambda n, m: n != m

        def canonicalBeforeR(fTp, tTp):
            return lambda n, m: Crank[n - 1] < Crank[m - 1]

        def canonicalAfterR(fTp, tTp):
            return lambda n, m: Crank[n - 1] > Crank[m - 1]

        def spinSameSlots(fTp, tTp):
            if fTp == slotType and tTp == slotType:

                def doyarns(yF, yT):
                    x = set(yF) & set(yT)
                    return (x, x)

                return doyarns
            elif fTp == slotType or tTp == slotType:

                def doyarns(yS, y2):
                    sindex = {}
                    for m in y2:
                        ss = Eoslots[m - maxSlot - 1]
                        if len(ss) == 1:
                            sindex.setdefault(ss[0], set()).add(m)
                    nyS = yS & set(sindex.keys())
                    ny2 = reduce(
                        set.union,
                        (sindex[s] for s in nyS),
                        set(),
                    )
                    return (nyS, ny2)

                if fTp == slotType:
                    return doyarns
                else:

                    def xx(yF, yT):
                        (nyT, nyF) = doyarns(yT, yF)
                        return (nyF, nyT)

                    return xx
            else:

                def doyarns(yF, yT):
                    sindexF = {}
                    for n in yF:
                        s = frozenset(Eoslots[n - maxSlot - 1])
                        sindexF.setdefault(s, set()).add(n)
                    sindexT = {}
                    for m in yT:
                        s = frozenset(Eoslots[m - maxSlot - 1])
                        sindexT.setdefault(s, set()).add(m)
                    nyS = set(sindexF.keys()) & set(sindexT.keys())
                    nyF = reduce(
                        set.union,
                        (sindexF[s] for s in nyS),
                        set(),
                    )
                    nyT = reduce(
                        set.union,
                        (sindexT[s] for s in nyS),
                        set(),
                    )
                    return (nyF, nyT)

                return doyarns

        def sameSlotsR(fTp, tTp):
            if fTp == slotType and tTp == slotType:
                return lambda n: (n, )
            elif tTp == slotType:
                return lambda n, m: Eoslots[n - maxSlot - 1] == (m, )
            elif fTp == slotType:
                return lambda n, m: Eoslots[m - maxSlot - 1] == (n, )
            else:
                return (
                    lambda n, m: (
                        frozenset(Eoslots[n-maxSlot-1]) ==
                        frozenset(Eoslots[m-maxSlot-1])
                    )
                )

        def spinOverlap(fTp, tTp):
            if fTp == slotType and tTp == slotType:

                def doyarns(yF, yT):
                    x = set(yF) & set(yT)
                    return (x, x)

                return doyarns
            elif fTp == slotType or tTp == slotType:

                def doyarns(yS, y2):
                    sindex = {}
                    for m in y2:
                        for s in Eoslots[m - maxSlot - 1]:
                            sindex.setdefault(s, set()).add(m)
                    nyS = yS & set(sindex.keys())
                    ny2 = reduce(
                        set.union,
                        (sindex[s] for s in nyS),
                        set(),
                    )
                    return (nyS, ny2)

                if fTp == slotType:
                    return doyarns
                else:

                    def xx(yF, yT):
                        (nyT, nyF) = doyarns(yT, yF)
                        return (nyF, nyT)

                    return xx
            else:

                def doyarns(yF, yT):
                    REDUCE_FACTOR = 0.4
                    SIZE_LIMIT = 10000
                    sindexF = {}
                    for n in yF:
                        for s in Eoslots[n - maxSlot - 1]:
                            sindexF.setdefault(s, set()).add(n)
                    sindexT = {}
                    for m in yT:
                        for s in Eoslots[m - maxSlot - 1]:
                            sindexT.setdefault(s, set()).add(m)
                    nyS = set(sindexF.keys()) & set(sindexT.keys())

                    lsF = len(sindexF)
                    lsT = len(sindexT)
                    lsI = len(nyS)
                    if lsF == lsT:  # spinning is completely useless
                        return (yF, yT)
                    if lsI > REDUCE_FACTOR * lsT and lsT > SIZE_LIMIT:
                        # spinning is not worth it
                        return (yF, yT)

                    if not self.silent:
                        self.api.info(
                            '1. reducing over {} elements'.format(len(nyS))
                        )
                    nyF = reduce(
                        set.union,
                        (sindexF[s] for s in nyS),
                        set(),
                    )
                    if not self.silent:
                        self.api.info(
                            '2. reducing over {} elements'.format(len(nyS))
                        )
                    nyT = reduce(
                        set.union,
                        (sindexT[s] for s in nyS),
                        set(),
                    )
                    return (nyF, nyT)

                return doyarns

        def overlapR(fTp, tTp):
            if fTp == slotType and tTp == slotType:
                return lambda n: (n, )
            elif tTp == slotType:
                return lambda n: Eoslots[n - maxSlot - 1]
            elif fTp == slotType:
                return lambda n, m: n in frozenset(Eoslots[m - maxSlot - 1])
            else:
                return (
                    lambda n, m: (
                        len(frozenset(Eoslots[n-maxSlot-1]) &
                            frozenset(Eoslots[m-maxSlot-1])) != 0
                    )
                )

        def diffSlotsR(fTp, tTp):
            if fTp == slotType and tTp == slotType:
                return lambda n, m: m != n
            elif tTp == slotType:
                return lambda n, m: Eoslots[m - maxSlot - 1] != (n, )
            elif fTp == slotType:
                return lambda n, m: Eoslots[n - maxSlot - 1] != (m, )
            else:
                return (
                    lambda n, m: (
                        frozenset(Eoslots[n-maxSlot-1]) !=
                        frozenset(Eoslots[m-maxSlot-1])
                    )
                )

        def disjointSlotsR(fTp, tTp):
            if fTp == slotType and tTp == slotType:
                return lambda n, m: m != n
            elif tTp == slotType:
                return lambda n, m: m not in frozenset(Eoslots[n-maxSlot-1])
            elif fTp == slotType:
                return lambda n, m: n not in frozenset(Eoslots[m-maxSlot-1])
            else:
                return (
                    lambda n, m: (
                        len(frozenset(Eoslots[n-maxSlot-1]) &
                            frozenset(Eoslots[m-maxSlot-1])) == 0
                    )
                )

        def inR(fTp, tTp):
            if fTp == slotType and tTp == slotType:
                return lambda n: ()
            elif tTp == slotType:
                return lambda n: ()
            elif fTp == slotType:
                return lambda n: ClevUp[n - 1]
            else:
                return lambda n: ClevUp[n - 1]

        def hasR(fTp, tTp):
            if fTp == slotType and tTp == slotType:
                return lambda n: ()
            elif fTp == slotType:
                return lambda n: ()
            elif tTp == slotType:
                return lambda n: Eoslots[n - maxSlot - 1]
            else:
                return lambda n: ClevDown[n - maxSlot - 1]

        def slotBeforeR(fTp, tTp):
            if fTp == slotType and tTp == slotType:
                return lambda n, m: n < m
            elif fTp == slotType:
                return lambda n, m: n < Eoslots[m - maxSlot - 1][0]
            elif tTp == slotType:
                return lambda n, m: Eoslots[n - maxSlot - 1][-1] < m
            else:
                return (
                    lambda n, m: (
                        Eoslots[n - maxSlot - 1][-1] <
                        Eoslots[m - maxSlot - 1][0]
                    )
                )

        def slotAfterR(fTp, tTp):
            if fTp == slotType and tTp == slotType:
                return lambda n, m: n > m
            elif fTp == slotType:
                return lambda n, m: n > Eoslots[m - maxSlot - 1][-1]
            elif tTp == slotType:
                return lambda n, m: Eoslots[n - maxSlot - 1][0] > m
            else:
                return (
                    lambda n, m: (
                        Eoslots[n - maxSlot - 1][0] >
                        Eoslots[m - maxSlot - 1][-1]
                    )
                )

        def sameFirstSlotR(fTp, tTp):
            if fTp == slotType and tTp == slotType:
                return lambda n: (n, )
            elif fTp == slotType:
                return lambda n: CfirstSlots[n - 1]
            elif tTp == slotType:
                return lambda n: (Eoslots[n - maxSlot - 1][0], )
            else:

                def xx(n):
                    fn = Eoslots[n - maxSlot - 1][0]
                    return CfirstSlots[fn - 1]

                return xx

        def sameLastSlotR(fTp, tTp):
            if fTp == slotType and tTp == slotType:
                return lambda n: (n, )
            elif fTp == slotType:
                return lambda n: ClastSlots[n - 1]
            elif tTp == slotType:
                return lambda n: (Eoslots[n - maxSlot - 1][-1], )
            else:

                def xx(n):
                    ln = Eoslots[n - maxSlot - 1][-1]
                    return ClastSlots[ln - 1]

                return xx

        def sameBoundaryR(fTp, tTp):
            if fTp == slotType and tTp == slotType:
                return lambda n: (n, )
            elif fTp == slotType:

                def xx(n):
                    fok = set(CfirstSlots[n - 1])
                    lok = set(ClastSlots[n - 1])
                    return tuple(fok & lok)

                return xx
            elif tTp == slotType:

                def xx(n):
                    slots = Eoslots[n - maxSlot - 1]
                    fs = slots[0]
                    ls = slots[-1]
                    return (fs, ) if fs == ls else ()

                return xx
            else:

                def xx(n):
                    fn = Eoslots[n - maxSlot - 1][0]
                    ln = Eoslots[n - maxSlot - 1][-1]
                    fok = set(CfirstSlots[fn - 1])
                    lok = set(ClastSlots[ln - 1])
                    return tuple(fok & lok)

                return xx

        def nearFirstSlotR(k):
            def zz(fTp, tTp):
                if fTp == slotType and tTp == slotType:
                    return lambda n: tuple(
                        m for m in range(max((1, n-k)), min((maxSlot, n+k+1)))
                    )
                elif fTp == slotType:

                    def xx(n):
                        near = set(
                            l
                            for l in
                            range(max((1, n - k)), min((maxSlot, n + k + 1)))
                        )
                        return tuple(
                            reduce(
                                set.union,
                                (set(CfirstSlots[l - 1]) for l in near),
                                set(),
                            )
                        )

                    return xx
                elif tTp == slotType:

                    def xx(n):
                        fn = Eoslots[n - maxSlot - 1][0]
                        return tuple(
                            m
                            for m in range(
                                max((1, fn - k)), min((maxSlot, fn + k + 1))
                            )
                        )

                    return xx
                else:

                    def xx(n):
                        fn = Eoslots[n - maxSlot - 1][0]
                        near = set(
                            l
                            for l in range(
                                max((1, fn - k)), min((maxSlot, fn + k + 1))
                            )
                        )
                        return tuple(
                            reduce(
                                set.union,
                                (set(CfirstSlots[l - 1]) for l in near),
                                set(),
                            )
                        )

                    return xx

            return zz

        def nearLastSlotR(k):
            def zz(fTp, tTp):
                if fTp == slotType and tTp == slotType:
                    return lambda n: tuple(
                        m for m in range(max((1, n-k)), min((maxSlot, n+k+1)))
                    )
                elif fTp == slotType:

                    def xx(n):
                        near = set(
                            l
                            for l in
                            range(max((1, n - k)), min((maxSlot, n + k + 1)))
                        )
                        return tuple(
                            reduce(
                                set.union,
                                (set(ClastSlots[l - 1]) for l in near),
                                set(),
                            )
                        )

                    return xx
                elif tTp == slotType:

                    def xx(n):
                        ln = Eoslots[n - maxSlot - 1][-1]
                        return tuple(
                            m
                            for m in range(
                                max((1, ln - k)), min((maxSlot, ln + k + 1))
                            )
                        )

                    return xx
                else:

                    def xx(n):
                        ln = Eoslots[n - maxSlot - 1][-1]
                        near = set(
                            l
                            for l in range(
                                max((1, ln - k)), min((maxSlot, ln + k + 1))
                            )
                        )
                        return tuple(
                            reduce(
                                set.union,
                                (set(ClastSlots[l - 1]) for l in near),
                                set(),
                            )
                        )

                    return xx

            return zz

        def nearBoundaryR(k):
            def zz(fTp, tTp):
                if fTp == slotType and tTp == slotType:
                    return lambda n: tuple(
                        m for m in range(max((1, n-k)), min((maxSlot, n+k+1)))
                    )
                elif fTp == slotType:

                    def xx(n):
                        near = set(
                            l
                            for l in
                            range(max((1, n - k)), min((maxSlot, n + k + 1)))
                        )
                        fok = set(
                            reduce(
                                set.union,
                                (set(CfirstSlots[l - 1]) for l in near),
                                set(),
                            )
                        )
                        lok = set(
                            reduce(
                                set.union,
                                (set(ClastSlots[l - 1]) for l in near),
                                set(),
                            )
                        )
                        return tuple(fok & lok)

                    return xx
                elif tTp == slotType:

                    def xx(n):
                        slots = Eoslots[n - maxSlot - 1]
                        fs = slots[0]
                        ls = slots[-1]
                        fok = set(
                            m
                            for m in range(
                                max((1, fs - k)), min((maxSlot, fs + k + 1))
                            )
                        )
                        lok = set(
                            m
                            for m in range(
                                max((1, ls - k)), min((maxSlot, ls + k + 1))
                            )
                        )
                        return tuple(fok & lok)

                    return xx
                else:

                    def xx(n):
                        fn = Eoslots[n - maxSlot - 1][0]
                        ln = Eoslots[n - maxSlot - 1][-1]
                        nearf = set(
                            ls
                            for ls in range(
                                max((1, fn - k)), min((maxSlot, fn + k + 1))
                            )
                        )
                        nearl = set(
                            ls
                            for ls in range(
                                max((1, ln - k)), min((maxSlot, ln + k + 1))
                            )
                        )
                        fok = set(CfirstSlots[fn - 1])
                        lok = set(ClastSlots[ln - 1])
                        fok = set(
                            reduce(
                                set.union,
                                (set(CfirstSlots[ls - 1]) for ls in nearf),
                                set(),
                            )
                        )
                        lok = set(
                            reduce(
                                set.union,
                                (set(ClastSlots[ls - 1]) for ls in nearl),
                                set(),
                            )
                        )
                        return tuple(fok & lok)

                    return xx

            return zz

        def adjBeforeR(fTp, tTp):
            if fTp == slotType and tTp == slotType:
                return lambda n: (n + 1, ) if n < maxSlot else ()
            else:

                def xx(n):
                    if n == maxSlot:
                        return ()
                    myNext = n + 1 if n < maxSlot else Eoslots[n - maxSlot -
                                                               1][-1] + 1
                    if myNext > maxSlot:
                        return ()
                    return CfirstSlots[myNext - 1] + (myNext, )

                return xx

        def adjAfterR(fTp, tTp):
            if fTp == slotType and tTp == slotType:
                return lambda n: (n - 1, ) if n > 1 else ()
            else:

                def xx(n):
                    if n <= 1:
                        return ()
                    myPrev = n - 1 if n < maxSlot else Eoslots[n - maxSlot -
                                                               1][0] - 1
                    if myPrev <= 1:
                        return ()
                    return (myPrev, ) + ClastSlots[myPrev - 1]

                return xx

        def nearBeforeR(k):
            def zz(fTp, tTp):
                if fTp == slotType and tTp == slotType:
                    return lambda n: tuple(
                        m for m in
                        range(max((1, n+1-k)), min((maxSlot, n+1+k+1)))
                    )
                else:

                    def xx(n):
                        myNext = n + 1 if n < maxSlot else Eoslots[n - maxSlot
                                                                   - 1][-1] + 1
                        myNextNear = tuple(
                            ls
                            for ls in range(
                                max((1, myNext - k
                                     )), min((maxSlot, myNext + k + 1))
                            )
                        )
                        nextSet = set(
                            reduce(
                                set.union,
                                (
                                    set(CfirstSlots[ls - 1])
                                    for ls in myNextNear
                                ),
                                set(),
                            )
                        )
                        return tuple(nextSet) + myNextNear

                    return xx

            return zz

        def nearAfterR(k):
            def zz(fTp, tTp):
                if fTp == slotType and tTp == slotType:
                    return lambda n: tuple(
                        m for m in
                        range(max((1, n-1-k)), min((maxSlot, n-1+k+1)))
                    )
                else:

                    def xx(n):
                        myPrev = n - 1 if n < maxSlot else Eoslots[n - maxSlot
                                                                   - 1][0] - 1
                        myPrevNear = tuple(
                            l
                            for l in range(
                                max((1, myPrev - k
                                     )), min((maxSlot, myPrev + k + 1))
                            )
                        )
                        prevSet = set(
                            reduce(
                                set.union,
                                (set(ClastSlots[l - 1]) for l in myPrevNear),
                                set(),
                            )
                        )
                        return tuple(prevSet) + myPrevNear

                    return xx

            return zz

        def makeEdgeMaps(efName):
            def edgeR(ftP, tTp):
                Es = self.api.Es
                return lambda n: Es(efName).f(n)

            def edgeIR(ftP, tTp):
                Es = self.api.Es
                return lambda n: Es(efName).t(n)

            return (edgeR, edgeIR)

        relations = [
            (
                ('=', spinEqual, equalR, 'left equal to right (as node)'),
                ('=', spinEqual, equalR, None),
            ),
            (
                ('#', 0.999, unequalR, 'left unequal to right (as node)'),
                ('#', 0.999, unequalR, None),
            ),
            (
                (
                    '<', 0.500, canonicalBeforeR,
                    'left before right (in canonical node ordering)'
                ),
                (
                    '>', 0.500, canonicalAfterR,
                    'left after right (in canonical node ordering)'
                ),
            ),
            (
                (
                    '==', spinSameSlots, sameSlotsR,
                    'left occupies same slots as right'
                ),
                ('==', spinSameSlots, sameSlotsR, None),
            ),
            (
                (
                    '&&', spinOverlap, overlapR,
                    'left has overlapping slots with right'
                ),
                ('&&', spinOverlap, overlapR, None),
            ),
            (
                (
                    '##', 0.990, diffSlotsR,
                    'left and right do not have the same slot set'
                ),
                ('##', 0.990, diffSlotsR, None),
            ),
            (
                (
                    '||', 0.900, disjointSlotsR,
                    'left and right do not have common slots'
                ),
                ('||', 0.900, disjointSlotsR, None),
            ),
            (
                ('[[', True, hasR, 'left embeds right'),
                (']]', True, inR, 'left embedded in right'),
            ),
            (
                ('<<', 0.490, slotBeforeR, 'left completely before right'),
                ('>>', 0.490, slotAfterR, 'left completely after right'),
            ),
            (
                (
                    '=:', True, sameFirstSlotR,
                    'left and right start at the same slot'
                ),
                ('=:', True, sameFirstSlotR, None),
            ),
            (
                (
                    ':=', True, sameLastSlotR,
                    'left and right end at the same slot'
                ),
                (':=', True, sameLastSlotR, None),
            ),
            (
                (
                    '::', True, sameBoundaryR,
                    'left and right start and end at the same slot'
                ),
                ('::', True, sameBoundaryR, None),
            ),
            (
                ('<:', True, adjBeforeR, 'left immediately before right'),
                (':>', True, adjAfterR, 'left immediately after right'),
            ),
            (
                (
                    '=k:', True, nearFirstSlotR,
                    'left and right start at k-nearly the same slot'
                ),
                ('=k:', True, nearFirstSlotR, None),
            ),
            (
                (
                    ':k=', True, nearLastSlotR,
                    'left and right end at k-nearly the same slot'
                ),
                (':k=', True, nearLastSlotR, None),
            ),
            (
                (
                    ':k:', True, nearBoundaryR,
                    'left and right start and end at k-near slots'
                ),
                (':k:', True, nearBoundaryR, None),
            ),
            (
                ('<k:', True, nearBeforeR, 'left k-nearly before right'),
                (':k>', True, nearAfterR, 'left k-nearly after right'),
            ),
        ]

        self.api.TF.explore(silent=self.silent)
        edgeMap = {}

        for efName in sorted(self.api.TF.featureSets['edges']):
            if efName == WARP[1]:
                continue
            r = len(relations)

            (edgeR, edgeIR) = makeEdgeMaps(efName)
            relations.append((
                (
                    '-{}>'.format(efName), True, edgeR,
                    'edge feature "{}"'.format(efName)
                ),
                (
                    '<{}-'.format(efName), True, edgeIR,
                    'edge feature "{}" (opposite direction)'.format(efName)
                ),
            ))
            edgeMap[2 * r] = (efName, 1)
            edgeMap[2 * r + 1] = (efName, -1)
        lr = len(relations)

        relationsAll = []
        for (r, rc) in relations:
            relationsAll.extend([r, rc])

        self.relations = [
            dict(
                acro=r[0],
                spin=r[1],
                func=r[2],
                desc=r[3],
            ) for r in relationsAll
        ]
        self.relationFromName = dict(((r['acro'], i)
                                      for (i, r) in enumerate(self.relations)))
        self.relationLegend = '\n'.join(
            '{:>23} {}'.format(r['acro'], r['desc']) for r in self.relations
            if r['desc'] is not None
        )
        self.relationLegend += '''
The warp feature "{}" cannot be used in searches.
One of the above relations on nodes and/or slots will suit you better!
'''.format(WARP[1])
        self.converse = dict(
            tuple((2 * i, 2 * i + 1)
                  for i in range(lr)) + tuple((2 * i + 1, 2 * i)
                                              for i in range(lr))
        )
        self.edgeMap = edgeMap

    def _addRelations(self, varRels):
        relations = self.relations
        tasks = collections.defaultdict(set)
        for (acro, ks) in varRels.items():
            j = self.relationFromName[acro]
            ji = self.converse[j]
            if ji < j:
                (j, ji) = (ji, j)
            acro = relations[j]['acro']
            acroi = relations[ji]['acro']
            tasks[(j, acro, ji, acroi)] |= ks

        for ((j, acro, ji, acroi), ks) in tasks.items():
            for k in ks:
                newAcro = acro.replace('k', str(k))
                newAcroi = acroi.replace('k', str(k))
                r = relations[j]
                ri = relations[ji]
                lr = len(relations)
                relations.extend([
                    dict(
                        acro=newAcro,
                        spin=r['spin'],
                        func=r['func'](k),
                        desc=r['desc'],
                    ),
                    dict(
                        acro=newAcroi,
                        spin=ri['spin'],
                        func=ri['func'](k),
                        desc=ri['desc'],
                    ),
                ])
                self.relationFromName[newAcro] = lr
                self.relationFromName[newAcroi] = lr + 1
                self.converse[lr] = lr + 1
                self.converse[lr + 1] = lr

# SYNTACTIC ANALYSIS OF SEARCH TEMPLATE ###

    def _tokenize(self):
        if not self.good:
            return

        opPat = '(?:[#&|\[\]<>:=-]+\S*)'
        atomPat = '(\s*)([^ \t=!~]+)(?:(?:\s*\Z)|(?:\s+(.*)))$'
        atomOpPat = (
            '(\s*)({op})\s+([^ \t=!~]+)(?:(?:\s*\Z)|(?:\s+(.*)))$'.format(
                op=opPat
            )
        )
        opLinePat = '(\s*)({op})\s*$'.format(op=opPat)
        namePat = '[A-Za-z0-9_-]+'
        relPat = '^\s*({nm})\s+({op})\s+({nm})\s*$'.format(
            nm=namePat, op=opPat
        )

        atomOpRe = re.compile(atomOpPat)
        atomRe = re.compile(atomPat)
        opLineRe = re.compile(opLinePat)
        nameRe = re.compile('^{}$'.format(namePat))
        relRe = re.compile(relPat)
        whiteRe = re.compile('^\s*$')

        nonePat = '^([a-zA-Z0-9-_]+)!\s*$'
        noneRe = re.compile(nonePat)

        def getFeatures(x, i):
            features = {}
            featureString = x.replace('\\ ', chr(1)) if x is not None else ''
            featureList = featureString.split()
            good = True
            for featStr in featureList:
                feat = featStr.replace(chr(1), ' ')
                match = noneRe.match(feat)
                if match:
                    (featN,) = match.groups()
                    featName = unesc(featN)
                    featVals = None
                else:
                    match = reRe.match(feat)
                    if match:
                        (featN, valRe) = match.groups()
                        featName = unesc(featN)
                        try:
                            featVals = re.compile(valRe)
                        except Exception() as err:
                            self.badSyntax.append(
                                'Wrong regular expression in line'
                                f'{i}: "{err}"'
                            )
                            good = False
                            featVals = None
                    else:
                        featComps = feat.split('=', 1)
                        featName = unesc(featComps[0])
                        if len(featComps) == 1:
                            featVals = True
                        else:
                            featValList = featComps[1]
                            featVals = set(
                                unesc(featVal)
                                for featVal in featValList.split('|')
                            )
                features[featName] = featVals
            return features if good else None

        searchLines = self.searchLines
        tokens = []
        allGood = True
        for (i, line) in enumerate(searchLines):
            if line.startswith('#') or whiteRe.match(line):
                continue
            good = False
            for x in [True]:
                match = opLineRe.match(line)
                if match:
                    (indent, op) = match.groups()
                    tokens.append((i, 'atom', len(indent), op))
                    good = True
                    break
                match = relRe.match(line)
                if match:
                    tokens.append((
                        i, 'rel', match.group(1), match.group(2),
                        match.group(3)
                    ))
                    good = True
                    break
                matchOp = atomOpRe.match(esc(line))
                if not matchOp:
                    match = atomRe.match(esc(line))
                if matchOp:
                    (indent, op, atom, features) = matchOp.groups()
                elif match:
                    op = None
                    (indent, atom, features) = match.groups()
                if matchOp or match:
                    atomComps = atom.split(':', 1)
                    if len(atomComps) == 1:
                        name = ''
                    else:
                        name = unesc(atomComps[0])
                        atom = unesc(atomComps[1])
                        mt = nameRe.match(name)
                        if not mt:
                            self.badSyntax.append(
                                'Illegal name at line {}: "{}"'.format(
                                    i, name
                                )
                            )
                            good = False
                    features = getFeatures(features, i)
                    if features is None:
                        good = False
                    else:
                        tokens.append(
                            (i, 'atom', len(indent), op, name, atom, features)
                        )
                        good = True
                    break
                features = getFeatures(esc(line), i)
                if features is None:
                    good = False
                else:
                    tokens.append((i, 'feat', features))
                    good = True
                break
            if not good:
                allGood = False
        if allGood:
            self.tokens = tokens
        else:
            self.good = False

    def _syntax(self):
        error = self.api.error
        self.good = True
        self.badSyntax = []
        self.searchLines = self.searchTemplate.split('\n')
        self._tokenize()
        if not self.good:
            for (i, line) in enumerate(self.searchLines):
                error('{:>2} {}'.format(i, line), tm=False)
            for eline in self.badSyntax:
                error(eline, tm=False)

# SEMANTIC ANALYSIS OF SEARCH TEMPLATE ###

    def _grammar(self):
        if not self.good:
            return

        prevKind = None
        good = True
        qnames = {}
        qnodes = []
        qedges = []
        edgeLine = {}
        nodeLine = {}
        tokens = sorted(
            self.tokens,
            key=lambda t: (len(self.tokens) + t[0]) if t[1] == 'rel' else t[0]
        )

        # atomStack is a stack of qnodes with their indent levels
        # such that every next member is one level deeper
        # and every member is the last qnode encountered at that level
        # The stack is implemented as a dict,
        # keyed by the indent, and valued by the qnode
        atomStack = {}

        for (i, kind, *fields) in tokens:
            if kind == 'atom':
                if len(fields) == 2:
                    (indent, op) = fields
                else:
                    (indent, op, name, otype, features) = fields
                    qnodes.append((otype, features))
                    q = len(qnodes) - 1
                    nodeLine[q] = i
                    name = ':{}'.format(i) if name == '' else name
                    qnames[name] = q
                if len(atomStack) == 0:
                    if indent > 0:
                        self.badSemantics.append(
                            'Unexpected indent at line {}: {}, expected {}'.
                            format(i, indent, 0)
                        )
                        good = False
                    if op is not None:
                        self.badSemantics.append((
                            'Lonely relation at line {}:'
                            ' not allowed at outermost level'
                        ).format(i))
                        good = False
                    if len(fields) > 2:
                        atomStack[0] = q
                else:
                    atomNest = sorted(atomStack.items(), key=lambda x: x[0])
                    top = atomNest[-1]
                    if indent == top[0]:
                        # sibling of previous atom
                        if len(atomNest) > 1:
                            if len(fields) == 2:
                                # lonely operator:
                                # left is previous atom, right is parent atom
                                qedges.append((top[1], op, atomNest[-2][1]))
                                edgeLine[len(qedges) - 1] = i
                            else:
                                # take the qnode of the subtop of the
                                # atomStack, if there is one
                                qedges.append((q, ']]', atomNest[-2][1]))
                                edgeLine[len(qedges) - 1] = i
                                if op is not None:
                                    qedges.append((top[1], op, q))
                                    edgeLine[len(qedges) - 1] = i
                    elif indent > top[0]:
                        if len(fields) == 2:
                            self.badSemantics.append((
                                'Lonely relation at line {}:'
                                ' not allowed as first child'
                            ).format(i))
                            good = False
                        else:
                            # child of previous atom
                            qedges.append((q, ']]', top[1]))
                            edgeLine[len(qedges) - 1] = i
                            if op is not None:
                                qedges.append((top[1], op, q))
                                edgeLine[len(qedges) - 1] = i
                    else:
                        # outdent action:
                        # look up the proper parent in the stack
                        if indent not in atomStack:
                            # parent cannot be found: indentation error
                            self.badSemantics.append((
                                'Unexpected indent at line {}:'
                                ' {}, expected one of {}'
                            ).format(
                                i,
                                indent,
                                ','.join(
                                    str(at[0]) for at in atomNest
                                    if at[0] < indent
                                ),
                            ))
                            good = False
                        else:
                            parents = [
                                at[1] for at in atomNest if at[0] < indent
                            ]
                            if len(
                                parents
                            ) != 0:  # if not already at outermost level
                                if len(
                                    fields
                                ) == 2:  # connect previous sibling to parent
                                    qedges.append(
                                        (atomStack[indent], op, parents[-1])
                                    )
                                    edgeLine[len(qedges) - 1] = i
                                else:
                                    qedges.append((q, ']]', parents[-1]))
                                    edgeLine[len(qedges) - 1] = i
                                    if op is not None:
                                        qedges.append(
                                            (atomStack[indent], op, q)
                                        )
                                        edgeLine[len(qedges) - 1] = i
                            removeKeys = [
                                at[0] for at in atomNest if at[0] > indent
                            ]
                            for rk in removeKeys:
                                del atomStack[rk]
                    atomStack[indent] = q
            elif kind == 'feat':
                features = fields[0]
                if prevKind is not None and prevKind != 'atom':
                    self.badSemantics.append(
                        'Features without atom at line {}: "{}"'.format(
                            i, features
                        )
                    )
                    good = False
                else:
                    qnodes[-1][1].update(features)
            elif kind == 'rel':
                (fName, opName, tName) = fields
                f = qnames.get(fName, None)
                t = qnames.get(tName, None)
                namesGood = True
                for (q, n) in ((f, fName), (t, tName)):
                    if q is None:
                        self.badSemantics.append(
                            'Relation with undefined name at line {}: "{}"'.
                            format(i, n)
                        )
                        namesGood = False
                if not namesGood:
                    good = False
                else:
                    qedges.append((f, opName, t))
                    edgeLine[len(qedges) - 1] = i
            prevKind = kind
        if good:
            self.qnames = qnames
            self.qnodes = qnodes
            self.qedgesRaw = qedges
            self.nodeLine = nodeLine
            self.edgeLine = edgeLine
        else:
            self.good = False

    def _validation(self):
        if not self.good:
            return

        levels = self.api.C.levels.data
        otypes = set(x[0] for x in levels)
        qnodes = self.qnodes
        nodeLine = self.nodeLine
        edgeMap = self.edgeMap

        edgeLine = self.edgeLine
        relationFromName = self.relationFromName

        # check the object types of atoms

        otypesGood = True
        for (q, (otype, xx)) in enumerate(qnodes):
            if otype not in otypes:
                self.badSemantics.append(
                    'Unknown object type in line {}: "{}"'.format(
                        nodeLine[q], otype
                    )
                )
                otypesGood = False
        if not otypesGood:
            self.badSemantics.append(
                'Valid object types are: {}'.format(
                    ','.join(x[0] for x in levels),
                )
            )
            self.good = False

        # check the feature names of feature specs
        # and check the types of their values

        missingFeatures = {}
        wrongValues = {}
        for (q, (xx, features)) in enumerate(qnodes):
            for (fName, values) in features.items():
                if fName not in self.api.TF.featureSets['nodes']:
                    missingFeatures.setdefault(fName, []).append(q)
                else:
                    requiredType = self.api.TF.features[fName].dataType
                    if values is True:
                        continue
                    elif values is None:
                        continue
                    elif isinstance(values, reTp):
                        if requiredType == 'int':
                            wrongValues.setdefault(fName, {}).setdefault(
                                values, []
                            ).append(q)
                    else:
                        valuesCast = set()
                        if requiredType == 'int':
                            for val in values:
                                try:
                                    valCast = int(val)
                                except Exception:
                                    valCast = val
                                    wrongValues.setdefault(
                                        fName, {}).setdefault(
                                            val, []
                                    ).append(q)
                                valuesCast.add(valCast)
                            features[fName] = valuesCast

        if len(missingFeatures):
            for (fName, qs) in sorted(missingFeatures.items()):
                self.badSemantics.append(
                    'Missing feature "{}" in line(s) {}'.format(
                        fName,
                        ','.join(str(nodeLine[q]) for q in qs),
                    )
                )
            self.good = False

        if len(wrongValues):
            for (fName, wrongs) in sorted(wrongValues.items()):
                self.badSemantics.append(
                    'Feature "{}" has wrong values:'.format(fName)
                )
                for (val, qs) in sorted(wrongs.items()):
                    self.badSemantics.append(
                        '    "{}" is not a number: line(s) {}'.format(
                            val,
                            ','.join(str(nodeLine[q]) for q in qs),
                        )
                    )
            self.good = False

        # check the relational operator token in edges
        # and replace them by an index
        # in the relations array of known relations
        qedges = []
        edgesGood = True
        kRe = re.compile('^([^0-9]*)([0-9]+)([^0-9]*)$')
        addRels = {}
        for (e, (f, opName, t)) in enumerate(self.qedgesRaw):
            match = kRe.findall(opName)
            if len(match):
                (pre, k, post) = match[0]
                opNameK = '{}{}{}'.format(pre, 'k', post)
                addRels.setdefault(opNameK, set()).add(int(k))
        self._addRelations(addRels)
        for (e, (f, opName, t)) in enumerate(self.qedgesRaw):
            match = kRe.findall(opName)
            if len(match):
                (pre, k, post) = match[0]
                opNameK = '{}{}{}'.format(pre, 'k', post)
            rela = relationFromName.get(opName, None)
            if rela is None:
                self.badSemantics.append(
                    'Unknown relation in line {}: "{}"'.format(
                        edgeLine[e], opName
                    )
                )
                edgesGood = False
            qedges.append((f, rela, t))
        if not edgesGood:
            self.badSemantics.append(
                'Allowed relations:\n{}'.format(self.relationLegend)
            )
            self.good = False
        self.qedges = qedges

        # determine which node and edge features are not yet loaded,
        # and load them
        eFeatsUsed = set()
        for (f, rela, t) in qedges:
            efName = edgeMap.get(rela, (None, ))[0]
            if efName is not None:
                eFeatsUsed.add(efName)
        nFeatsUsed = set()
        for (xx, features) in qnodes:
            for nfName in features:
                nFeatsUsed.add(nfName)

        if self.good:
            needToLoad = []
            for fName in sorted(eFeatsUsed | nFeatsUsed):
                fObj = self.api.TF.features[fName]
                if not fObj.dataLoaded or not hasattr(self.api.F, fName):
                    needToLoad.append((fName, fObj))
            if len(needToLoad):
                self.api.TF.load(
                    [x[0] for x in needToLoad], add=True, silent=True,
                )

    def _semantics(self):
        self.badSemantics = []
        if not self.good:
            return
        error = self.api.error

        self._grammar()
        if not self.good:
            for (i, line) in enumerate(self.searchLines):
                error('{:>2} {}'.format(i, line), tm=False)
            for eline in self.badSemantics:
                error(eline, tm=False)
            return
        self._validation()
        if not self.good:
            for (i, line) in enumerate(self.searchLines):
                error('{:>2} {}'.format(i, line), tm=False)
            for eline in self.badSemantics:
                error(eline, tm=False)

# WORKING WITH THE SEARCH GRAPH ###

    def _connectedness(self):
        error = self.api.error
        qnodes = self.qnodes
        qedges = self.qedges

        componentIndex = dict(((q, {q}) for q in range(len(qnodes))))
        for (f, rela, t) in qedges:
            if f != t:
                componentIndex[f] |= componentIndex[t]
                for u in componentIndex[f] - {f}:
                    componentIndex[u] = componentIndex[f]
        components = sorted(set(frozenset(c) for c in componentIndex.values()))
        componentIndex = {}
        for c in components:
            for q in c:
                componentIndex[q] = c
        componentEdges = {}
        for (e, (f, rela, t)) in enumerate(qedges):
            c = componentIndex[f]
            componentEdges.setdefault(c, []).append(e)
        self.components = []
        for c in components:
            self.components.append([sorted(c), componentEdges.get(c, [])])
        lComps = len(self.components)
        if lComps == 0:
            error('Search without instructions. Tell me what to look for.')
            self.good = False
        elif lComps > 1:
            error(
                'More than one connected components ({}):'.format(
                    len(self.components)
                )
            )
            error(
                'Either run the subqueries one by one,'
                ' or connect the components by a relation',
                tm=False
            )
            self.good = False

    def _showNode(self, q, pos2=False):
        qnodes = self.qnodes
        yarns = self.yarns
        space = ' ' * 19
        nodeInfo = 'node {} {:>2}-{:<13} ({:>6}   choices)'.format(
            space,
            q,
            qnodes[q][0],
            len(yarns[q]),
        ) if pos2 else 'node {:>2}-{:<13} {} ({:>6}   choices)'.format(
            q,
            qnodes[q][0],
            space,
            len(yarns[q]),
        )
        self.api.info(nodeInfo, tm=False)

    def _showEdge(self, e, dir):
        qnodes = self.qnodes
        qedges = self.qedges
        converse = self.converse
        relations = self.relations
        spreads = self.spreads
        spreadsC = self.spreadsC
        (f, rela, t) = qedges[e]
        if dir == -1:
            (f, rela, t) = (t, converse[rela], f)
        self.api.info(
            'edge {:>2}-{:<13} {:^2} {:>2}-{:<13} ({:8.1f} choices)'.format(
                f,
                qnodes[f][0],
                relations[rela]['acro'],
                t,
                qnodes[t][0],
                spreads.get(e, -1) if dir == 1 else spreadsC.get(e, -1),
            ),
            tm=False
        )

    def _showYarns(self):
        for q in range(len(self.qnodes)):
            self._showNode(q)

# SPINNING ###

    def _spinAtom(self, q):
        F = self.api.F
        Fs = self.api.Fs
        qnodes = self.qnodes

        (otype, features) = qnodes[q]
        featureList = sorted(features.items())
        yarn = set()
        for n in F.otype.s(otype):
            good = True
            for (ft, val) in featureList:
                fval = Fs(ft).v(n)
                if val is None:
                    if fval is not None:
                        good = False
                        break
                elif val is True:
                    if fval is None:
                        good = False
                        break
                elif type(val) is str or type(val) is int:
                    if fval != val:
                        good = False
                        break
                elif isinstance(val, reTp):
                    if fval is None or not val.search(fval):
                        good = False
                        break
                else:
                    if fval not in val:
                        good = False
                        break
            if good:
                yarn.add(n)
        self.yarns[q] = yarn

    def _spinAtoms(self):
        qnodes = self.qnodes
        for q in range(len(qnodes)):
            self._spinAtom(q)

    def _estimateSpreads(self, both=False):
        TRY_LIMIT_F = 10
        TRY_LIMIT_T = 10
        qnodes = self.qnodes
        relations = self.relations
        converse = self.converse
        qedges = self.qedges
        yarns = self.yarns

        self.spreadsC = {}
        self.spreads = {}

        for (e, (f, rela, t)) in enumerate(qedges):
            tasks = [(f, rela, t, 1)]
            if both:
                tasks.append((t, converse[rela], f, -1))
            for (tf, trela, tt, dir) in tasks:
                s = relations[trela]['spin']
                yarnF = yarns[tf]
                yarnT = yarns[tt]
                dest = self.spreads if dir == 1 else self.spreadsC
                if type(s) is float:
                    # fixed estimates
                    dest[e] = len(yarnT) * s
                    continue
                yarnF = list(yarnF)
                yarnT = yarns[tt]
                totalSpread = 0
                yarnFl = len(yarnF)
                if yarnFl < TRY_LIMIT_F:
                    triesn = yarnF
                else:
                    triesn = set(
                        yarnF[randrange(yarnFl)] for n in range(TRY_LIMIT_F)
                    )
                if len(triesn) == 0:
                    dest[e] = 0
                else:
                    r = relations[trela]['func'](qnodes[tf][0], qnodes[tt][0])
                    nparams = len(signature(r).parameters)
                    if nparams == 1:
                        for n in triesn:
                            mFromN = set(r(n)) & yarnT
                            totalSpread += len(mFromN)
                    else:
                        for n in triesn:
                            thisSpread = 0
                            yarnTl = len(yarnT)
                            if yarnTl < TRY_LIMIT_T:
                                triesm = yarnT
                            else:
                                triesm = set(
                                    list(yarnT)[randrange(yarnTl)]
                                    for m in range(TRY_LIMIT_T)
                                )
                            if len(triesm) == 0:
                                thisSpread = 0
                            else:
                                for m in triesm:
                                    if r(n, m):
                                        thisSpread += 1
                            totalSpread += yarnTl * thisSpread / len(triesm)
                    dest[e] = totalSpread / len(triesn)

    def _chooseEdge(self):
        F = self.api.F
        yarnFractionNode = {}
        qnodes = self.qnodes
        qedges = self.qedges
        spreads = self.spreads
        for (q, (otype, xx)) in enumerate(qnodes):
            (begin, end) = F.otype.sInterval(otype)
            nOtype = 1 + end - begin
            nYarn = len(self.yarns[q])
            yf = nYarn / nOtype
            yarnFractionNode[q] = yf * yf
        yarnFractionEdge = {}
        for (e, (f, rela, t)) in enumerate(qedges):
            if self.uptodate[e]:
                continue
            yarnFractionEdge[
                e
            ] = yarnFractionNode[f] + yarnFractionNode[t] + spreads[e]
        firstEdge = sorted(yarnFractionEdge.items(), key=lambda x: x[1])[0][0]
        return firstEdge

    def _spinEdge(self, e):
        SPIN_LIMIT = 1000
        qnodes = self.qnodes
        relations = self.relations
        yarns = self.yarns
        spreads = self.spreads
        qedges = self.qedges
        uptodate = self.uptodate

        (f, rela, t) = qedges[e]
        yarnF = yarns[f]
        yarnT = yarns[t]
        uptodate[e] = True

        # if the yarns around an edge are big,
        # and the spread of the relation is
        # also big, spinning costs an enormous amount of time,
        # and will not help in reducing the search space.
        # condition for skipping: spread times length from-yarn >= SPIN_LIMIT
        if spreads[e] * len(yarnF) >= SPIN_LIMIT:
            return

        # for some basic relations we know that spinning is useless
        s = relations[rela]['spin']
        if type(s) is float:
            return

        # for other basic realtions we have an optimized spin function
        # if type(s) is types.FunctionType:
        if isinstance(s, types.FunctionType):
            (newYarnF, newYarnT) = s(qnodes[f][0], qnodes[t][0])(yarnF, yarnT)
        else:
            r = relations[rela]['func'](qnodes[f][0], qnodes[t][0])
            nparams = len(signature(r).parameters)
            newYarnF = set()
            newYarnT = set()

            if nparams == 1:
                for n in yarnF:
                    mFromN = set(r(n)) & yarnT
                    if len(mFromN):
                        newYarnT |= mFromN
                        newYarnF.add(n)
            else:
                for n in yarnF:
                    mFromN = set(m for m in yarnT if r(n, m))
                    if len(mFromN):
                        newYarnT |= mFromN
                        newYarnF.add(n)

        affectedF = len(newYarnF) != len(yarns[f])
        affectedT = len(newYarnT) != len(yarns[t])

        uptodate[e] = True
        for (oe, (of, orela, ot)) in enumerate(qedges):
            if oe == e:
                continue
            if (affectedF and f in {of, ot}) or (affectedT and t in {of, ot}):
                self.uptodate[oe] = False
        self.yarns[f] = newYarnF
        self.yarns[t] = newYarnT

    def _spinEdges(self):
        qnodes = self.qnodes
        qedges = self.qedges
        yarns = self.yarns
        uptodate = self.uptodate

        self._estimateSpreads()

        for e in range(len(qedges)):
            uptodate[e] = False
        it = 0
        while 1:
            if min(len(yarns[q]) for q in range(len(qnodes))) == 0:
                break
            # if reduce(
            #    lambda y,z: y and z,
            #    (uptodate[e] for e in range(len(qedges))),
            #    True,
            # ): break
            if all(uptodate[e] for e in range(len(qedges))):
                break
            e = self._chooseEdge()
            (f, rela, t) = qedges[e]
            self._spinEdge(e)
            it += 1

# STITCHING: STRATEGIES ###

    def _spread_1_first(self):
        qedges = self.qedges
        qnodes = self.qnodes

        s1Edges = []
        for (e, (f, rela, t)) in enumerate(qedges):
            if self.spreads[e] <= 1:
                s1Edges.append((e, 1))
            if self.spreadsC[e] <= 1:
                s1Edges.append((e, -1))
# s1Edges contain all edges with spread <= 1, or whose converse has spread <= 1
# now we want to build the largest graph
# with the original nodes and these edges,
# such that you can walk from a starting point
# over directed s1 edges to every other point
# we initialize candidate graphs: for each node: singletons graph, no edges.
        candidates = []
        # add s1 edges and nodes to all candidates
        for q in range(len(qnodes)):
            cnodes = {q}
            cedges = set()
            cedgesOrder = []
            while 1:
                added = False
                for (e, dir) in s1Edges:
                    (f, rela, t) = qedges[e]
                    if dir == -1:
                        (t, f) = (f, t)
                    if f in cnodes:
                        if t not in cnodes:
                            cnodes.add(t)
                            added = True
                        if (e, dir) not in cedges:
                            cedges.add((e, dir))
                            cedgesOrder.append((e, dir))
                            added = True
                if not added:
                    break
            candidates.append((cnodes, cedgesOrder))

# pick the biggest graph (nodes and edges count for 1)
        startS1 = sorted(candidates, key=lambda x: len(x[0]) + len(x[1]))[-1]

        newNodes = startS1[0]
        newEdges = startS1[1]
        doneEdges = set(e[0] for e in newEdges)

        # we add all edges that are not yet in our startS1.
        # we add them two-fold: also with converse,
        # and we sort the result by spread
        # then we start a big loop:
        # in every iteration, we take the edge with smallest spread
        # that can be connected
        # to the graph under construction
        # then we start a new iteration, because the graph has grown,
        # and and new edges might
        # have become connectable by that

        # in order to fail early, we can also add edges
        # if their from-nodes and to-nodes both have been
        # targeted.
        # That means: an earlier edge went to f,
        # an other earlier edge went to t, and if we
        # have an edge from f to t, we'd better add it now,
        # since it is an extra constraint
        # and by testing it here we can avoid a lot of work.

        remainingEdges = set()
        for e in range(len(qedges)):
            if e not in doneEdges:
                remainingEdges.add((e, 1))
                remainingEdges.add((e, -1))
        remainingEdgesO = sorted(
            remainingEdges,
            key=lambda e: (
                self.spreads[e[0]] if e[1] == 1 else self.spreadsC[e[0]]
            ),
        )

        while 1:
            added = False
            for (e, dir) in remainingEdgesO:
                if e in doneEdges:
                    continue
                (f, rela, t) = qedges[e]
                if dir == -1:
                    (f, t) = (t, f)
                if f in newNodes and t in newNodes:
                    newEdges.append((e, dir))
                    doneEdges.add(e)
                    added = True
            for (e, dir) in remainingEdgesO:
                if e in doneEdges:
                    continue
                (f, rela, t) = qedges[e]
                if dir == -1:
                    (f, t) = (t, f)
                if f in newNodes:
                    newNodes.add(t)
                    newEdges.append((e, dir))
                    doneEdges.add(e)
                    added = True
                    break
            if not added:
                break

        self.newNodes = newNodes
        self.newEdges = newEdges

    def _small_choice_first(self):

        # This strategy does not try to make a big subgraph of
        # edges with spread 1.
        # The problem is that before the edges work,
        # the initial yarn may have an enormous spread.
        # Here we try out the strategy of postponing
        # broad choices as long as possible.
        # The inituition is that while we are making smaller choices,
        # constraints are encountered,
        # severely limiting the broader choices later on.

        # So, we pick the yarn with the least amount of nodes
        # as our starting point.
        # The corresponding node is our singleton start set.
        # In every iteration we do the following:
        # - we pick all edges of which from- and to-nodes
        #   are already in the node set
        # - we pick the edge with least spread
        #   that has a starting point in the set
        # Until nothing changes anymore

        qedges = self.qedges
        qnodes = self.qnodes

        newNodes = {
            sorted(range(len(qnodes)), key=lambda x: len(self.yarns[x]))[0]
        }
        newEdges = []
        doneEdges = set()

        remainingEdges = set()
        for e in range(len(qedges)):
            remainingEdges.add((e, 1))
            remainingEdges.add((e, -1))
        remainingEdgesO = sorted(
            remainingEdges,
            key=lambda e: (
                self.spreads[e[0]] if e[1] == 1 else self.spreadsC[e[0]]
            ),
        )

        while 1:
            added = False
            for (e, dir) in remainingEdgesO:
                if e in doneEdges:
                    continue
                (f, rela, t) = qedges[e]
                if dir == -1:
                    (f, t) = (t, f)
                if f in newNodes and t in newNodes:
                    newEdges.append((e, dir))
                    doneEdges.add(e)
                    added = True
            for (e, dir) in remainingEdgesO:
                if e in doneEdges:
                    continue
                (f, rela, t) = qedges[e]
                if dir == -1:
                    (f, t) = (t, f)
                if f in newNodes:
                    newNodes.add(t)
                    newEdges.append((e, dir))
                    doneEdges.add(e)
                    added = True
                    break
            if not added:
                break

        self.newNodes = newNodes
        self.newEdges = newEdges

    def _big_choice_first(self):

        # For comparison: the opposite of _small_choice_first.
        # Just to see what the performance difference is.

        qedges = self.qedges
        qnodes = self.qnodes

        newNodes = {
            sorted(range(len(qnodes)), key=lambda x: -len(self.yarns[x]))[0]
        }
        newEdges = []
        doneEdges = set()

        remainingEdges = set()
        for e in range(len(qedges)):
            remainingEdges.add((e, 1))
            remainingEdges.add((e, -1))
        remainingEdgesO = sorted(
            remainingEdges,
            key=lambda e: (
                -self.spreads[e[0]] if e[1] == 1 else -self.spreadsC[e[0]]
            ),
        )

        while 1:
            added = False
            for (e, dir) in remainingEdgesO:
                if e in doneEdges:
                    continue
                (f, rela, t) = qedges[e]
                if dir == -1:
                    (f, t) = (t, f)
                if f in newNodes and t in newNodes:
                    newEdges.append((e, dir))
                    doneEdges.add(e)
                    added = True
            for (e, dir) in remainingEdgesO:
                if e in doneEdges:
                    continue
                (f, rela, t) = qedges[e]
                if dir == -1:
                    (f, t) = (t, f)
                if f in newNodes:
                    newNodes.add(t)
                    newEdges.append((e, dir))
                    doneEdges.add(e)
                    added = True
                    break
            if not added:
                break

        self.newNodes = newNodes
        self.newEdges = newEdges

# STITCHING ###

    def _stitch(self):
        self._estimateSpreads(both=True)
        self._stitchPlan()
        if not self.good:
            return
        self._stitchResults()

# STITCHING: PLANNING ###

    def _stitchPlan(self, strategy=None):
        qnodes = self.qnodes
        qedges = self.qedges
        error = self.api.error

        self._setStrategy(strategy, keep=True)
        if not self.good:
            return

        # Apply the chosen strategy
        self.strategy()

        # remove spurious edges:
        # if we have both the 1 and -1 version of an edge,
        # we can leave out the one that we encounter in the second place

        newNodes = self.newNodes
        newEdges = self.newEdges

        newCedges = set()
        newCedgesOrder = []
        for (e, dir) in newEdges:
            if e not in newCedges:
                newCedgesOrder.append((e, dir))
                newCedges.add(e)

# conjecture: we have all edges and all nodes now
# reason: we work in a connected component, so all nodes are reachable
# by edges or inverses
# we check nevertheless

        qnodesO = tuple(range(len(qnodes)))
        newNodesO = tuple(sorted(newNodes))
        if newNodesO != qnodesO:
            error(
                '''Object mismatch in plan:
In template: {}
In plan    : {}'''.format(qnodesO, newNodesO),
                tm=False
            )
            self.good = False

        qedgesO = tuple(range(len(qedges)))
        newCedgesO = tuple(sorted(newCedges))
        if newCedgesO != qedgesO:
            error(
                '''Relation mismatch in plan:
In template: {}
In plan    : {}'''.format(qedgesO, newCedgesO),
                tm=False
            )
            self.good = False

        self.stitchPlan = (newNodes, newCedgesOrder)

# STITCHING: DELIVERING ###

    def _stitchResults(self):
        qnodes = self.qnodes
        qedges = self.qedges
        plan = self.stitchPlan
        relations = self.relations
        converse = self.converse
        yarns = self.yarns

        planEdges = plan[1]
        if len(planEdges) == 0:
            # no edges, hence a single node (because of connectedness,
            # hence we must deliver everything of its yarn
            yarn = yarns[0]

            def deliver(remap=True):
                for n in yarn:
                    yield (n, )

            self.results = deliver
            return

# The next function must be optimized, and the lookup of functions and data
# should be as direct as possible.
# Because deliver() below fetches the results,
# of wich there are unpredictably many.

# We are going to build-up and deliver stitches,
# which are instantiations of all the query nodes
# by text nodes in a specific sequence
# which is the same for all stitches.
# We can compile stitching in such a way, that the stitcher thinks it is
# instantiating q node 0, then 1, and so on.
# I.e. we are going to permute every thing that the stitching process sees,
# so that it happens in this order.

# We build up the stitch in a recursive process.
# When there is choice between a and b, we essentially say
#
# def build(stitch)
#     if there is choice
#        build(stitch+a)
#        build(stitch+b)
#
# But we do not have to pass on the stitch as an immutable data structure.
# We can just keep it as one single mutable datastructure, provided we
# do something between the two recursive calls above.
# Suppose stitch is an array, and in the outer build n elements are filled
# (the rest contains -1)
#
# Then we say
#     if there is choice
#        build(stitch+a)
#        for k in range(n, len(stitch)): stitch[k] = -1
#        build(stitch+b)
#
# It turns out that the data in stitch that is shared between calls
# is not modified by them.
# The only thing that happens, is that -1 values get new values.
# So coming out calls only requires us to restore -1's.
# And if the stitch is ordered in the right way,
# the -1's are always at the end.

# We start compiling and permuting

        edgesCompiled = []
        qPermuted = [
        ]  # row of nodes in the order as will be created during stitching
        qPermutedInv = {
        }  # mapping from original q node number to index in the permuted order
        for (i, (e, dir)) in enumerate(planEdges):
            (f, rela, t) = qedges[e]
            if dir == -1:
                (f, rela, t) = (t, converse[rela], f)
            r = relations[rela]['func'](qnodes[f][0], qnodes[t][0])
            nparams = len(signature(r).parameters)
            if i == 0:
                qPermuted.append(f)
                qPermutedInv[f] = len(qPermuted) - 1
            if t not in qPermuted:
                qPermuted.append(t)
                qPermutedInv[t] = len(qPermuted) - 1

            edgesCompiled.append(
                (qPermutedInv[f], qPermutedInv[t], r, nparams)
            )

# now permute the yarns

        yarnsPermuted = [yarns[q] for q in qPermuted]

        def deliver(remap=True):
            stitch = [None for q in range(len(qPermuted))]
            lStitch = len(stitch)
            edgesC = edgesCompiled
            yarnsP = yarnsPermuted

            def stitchOn(e):
                if e >= len(edgesC):
                    if remap:
                        yield tuple(
                            stitch[qPermutedInv[q]] for q in range(lStitch)
                        )
                    else:
                        yield tuple(stitch)
                    return
                (f, t, r, nparams) = edgesC[e]
                yarnT = yarnsP[t]
                if e == 0 and stitch[f] is None:
                    yarnF = yarnsP[f]
                    for sN in yarnF:
                        stitch[f] = sN
                        for s in stitchOn(e):
                            yield s
                    return
                sN = stitch[f]
                sM = stitch[t]
                if sM is not None:
                    if nparams == 1:
                        if sM in set(r(sN)):  # & yarnT:
                            for s in stitchOn(e + 1):
                                yield s
                    else:
                        if r(sN, sM):
                            for s in stitchOn(e + 1):
                                yield s
                    return
                mFromN = tuple(set(r(sN)) & yarnT) if nparams == 1 else tuple(
                    m for m in yarnT if r(sN, m)
                )
                for m in mFromN:
                    stitch[t] = m
                    for s in stitchOn(e + 1):
                        yield s
                stitch[t] = None

            for s in stitchOn(0):
                yield s

        self.results = deliver


# TOP-LEVEL IMPLEMENTATION METHODS

    def _parse(self):
        self._syntax()
        self._semantics()

    def _prepare(self):
        if not self.good:
            return
        self.yarns = {}
        self.spreads = {}
        self.spreadsC = {}
        self.uptodate = {}
        self.results = None
        self._connectedness()

    def _setStrategy(self, strategy, keep=False):
        error = self.api.error
        if strategy is None:
            if keep:
                return
            strategy = STRATEGY[0]
        if strategy not in STRATEGY:
            error('Strategy not defined: "{}"'.format(strategy))
            error(
                'Allowed strategies:\n{}'.format(
                    '\n'.join('    {}'.format(s) for s in STRATEGY)
                ),
                tm=False
            )
            self.good = False
            return

        method = '_{}'.format(strategy)
        if not hasattr(self, method):
            error(
                'Strategy is defined, but not implemented: "{}"'.
                format(strategy)
            )
            self.good = False
            return
        self.strategy = getattr(self, method)



import os
import pickle
import gzip
import collections
import time
from datetime import datetime
from .helpers import (
    setFromSpec, valueFromTf, tfFromValue, specFromRanges, rangesFromSet
)

ERROR_CUTOFF = 20
GZIP_LEVEL = 2
PICKLE_PROTOCOL = 4

WARP = (
    'otype',
    'oslots',
    'otext',
)

DATA_TYPES = ('str', 'int')


class Data(object):
    def __init__(
        self,
        path,
        tm,
        edgeValues=False,
        data=None,
        isEdge=None,
        isConfig=None,
        metaData={},
        method=None,
        dependencies=None
    ):
        (dirName, baseName) = os.path.split(path)
        (fileName, extension) = os.path.splitext(baseName)
        self.path = path
        self.tm = tm
        self.dirName = dirName
        self.fileName = fileName
        self.extension = extension
        self.binDir = '{}/.tf'.format(dirName)
        self.binPath = '{}/{}.tfx'.format(self.binDir, self.fileName)
        self.edgeValues = edgeValues
        self.isEdge = isEdge
        self.isConfig = isConfig
        self.metaData = metaData
        self.method = method
        self.dependencies = dependencies
        self.data = data
        self.dataLoaded = False
        self.dataError = False
        self.dataType = 'str'

    def load(self, metaOnly=False, silent=False):
        self.tm.indent(level=1, reset=True)
        origTime = self._getModified()
        binTime = self._getModified(bin=True)
        sourceRep = ', '.join(
            dep.fileName for dep in self.dependencies
        ) if self.method else self.dirName
        msgFormat = '{:<1} {:<20} from {}'
        actionRep = ''
        good = True

        if self.dataError:
            # there has been an error in an earlier
            # computation/compiling/loading of this feature
            actionRep = 'E'
            good = False
        elif self.dataLoaded and (
            self.isConfig or (not origTime or self.dataLoaded >= origTime) and
            (not binTime or self.dataLoaded >= binTime)
        ):
            actionRep = '='  # loaded and up to date
        elif not origTime and not binTime:
            actionRep = 'X'  # no source and no binary present
            good = False
        else:
            if not origTime:
                actionRep = 'b'
                good = self._readDataBin()
            elif not binTime or origTime > binTime:
                actionRep = 'C' if self.method else 'T'
                good = self._compute() if self.method else self._readTf(
                    metaOnly=metaOnly
                )
                if good:
                    if self.isConfig or metaOnly:
                        actionRep = 'M'
                    else:
                        self._writeDataBin()
            else:
                actionRep = 'B'
                good = True if self.method else self._readTf(metaOnly=True)
                if good:
                    if self.isConfig or metaOnly:
                        actionRep = 'M'
                    else:
                        good = self._readDataBin()
        if self.isConfig:
            self.cleanDataBin()
        if good:
            if actionRep != '=':
                self.tm.info(
                    msgFormat.format(actionRep, self.fileName, sourceRep),
                    cache=1 if (not silent) or (actionRep in 'CT') else -1,
                )
        else:
            self.dataError = True
            self.tm.error(
                msgFormat.format(actionRep, self.fileName, sourceRep)
            )
        return good

    def unload(self):
        self.data = None
        self.dataLoaded = False

    def save(self, overwrite=False, nodeRanges=False):
        return self._writeTf(overwrite=overwrite, nodeRanges=nodeRanges)

    def _setDataType(self):
        if self.isConfig:
            return
        if 'valueType' in self.metaData:
            dataType = self.metaData['valueType']
            if dataType not in DATA_TYPES:
                self.tm.error(
                    'Unknown @valueType: "{}". Should be one of {}'.format(
                        dataType,
                        ','.join(DATA_TYPES),
                    )
                )
                self.dataType = DATA_TYPES[0]
            else:
                self.dataType = dataType
        else:
            self.tm.error(
                'Missing @valueType. Should be one of {}'.format(
                    ','.join(DATA_TYPES),
                )
            )
            self.dataType = DATA_TYPES[0]

    def _readTf(self, metaOnly=False):
        path = self.path
        if not os.path.exists(path):
            self.tm.error(
                'TF reading: feature file "{}" does not exist'.format(path)
            )
            return False
        fh = open(path, encoding='utf8')
        i = 0
        self.metaData = {}
        self.isConfig = False
        for line in fh:
            i += 1
            if i == 1:
                text = line.rstrip()
                if text == '@edge':
                    self.isEdge = True
                elif text == '@node':
                    self.isEdge = False
                elif text == '@config':
                    self.isConfig = True
                else:
                    self.tm.error(
                        'Line {}: missing @node/@edge/@config'.format(i)
                    )
                    fh.close()
                    return False
                continue
            text = line.rstrip('\n')
            if len(text) and text[0] == '@':
                if text == '@edgeValues':
                    self.edgeValues = True
                    continue
                fields = text[1:].split('=', 1)
                self.metaData[fields[0]
                              ] = fields[1] if len(fields) == 2 else None
                continue
            else:
                if text != '':
                    self.tm.error(
                        'Line {}: missing blank line after metadata'.format(i)
                    )
                    fh.close()
                    return False
                else:
                    break
        self._setDataType()
        good = True
        if not metaOnly and not self.isConfig:
            good = self._readDataTf(fh, i)
        fh.close()
        return good

    def _readDataTf(self, fh, firstI):
        errors = collections.defaultdict(list)
        i = firstI
        implicit_node = 1
        data = {}
        isEdge = self.isEdge
        edgeValues = self.edgeValues
        normFields = 3 if isEdge and edgeValues else 2
        isNum = self.dataType == 'int'
        for line in fh:
            i += 1
            fields = line.rstrip('\n').split('\t')
            lfields = len(fields)
            if lfields > normFields:
                errors['wrongFields'].append(i)
                continue
            if lfields == normFields:
                nodes = setFromSpec(fields[0])
                if isEdge:
                    if fields[1] == '':
                        errors['emptyNode2Spec'].append(i)
                        continue
                    nodes2 = setFromSpec(fields[1])
                if not isEdge or edgeValues:
                    valTf = fields[-1]
            else:
                if isEdge:
                    if edgeValues:
                        if lfields == normFields - 1:
                            nodes = {implicit_node}
                            nodes2 = setFromSpec(fields[0])
                            valTf = fields[-1]
                        elif lfields == normFields - 2:
                            nodes = {implicit_node}
                            if fields[0] == '':
                                errors['emptyNode2Spec'].append(i)
                                continue
                            nodes2 = setFromSpec(fields[0])
                            valTf = ''
                        else:
                            nodes = {implicit_node}
                            valTf = ''
                            errors['emptyNode2Spec'].append(i)
                            continue
                    else:
                        if lfields == normFields - 1:
                            nodes = {implicit_node}
                            if fields[0] == '':
                                errors['emptyNode2Spec'].append(i)
                                continue
                            nodes2 = setFromSpec(fields[0])
                        else:
                            nodes = {implicit_node}
                            errors['emptyNode2Spec'].append(i)
                            continue
                else:
                    nodes = {implicit_node}
                    if lfields == 1:
                        valTf = fields[0]
                    else:
                        valTf = ''
            implicit_node = max(nodes) + 1
            if not isEdge or edgeValues:
                value = (
                    int(valTf) if isNum and valTf != '' else None
                    if isNum else '' if valTf == '' else valueFromTf(valTf)
                )
            if isEdge:
                for n in nodes:
                    for m in nodes2:
                        if not edgeValues:
                            data.setdefault(n, set()).add(m)
                        else:
                            data.setdefault(
                                n, {}
                            )[m] = value  # even if the value is None
            else:
                for n in nodes:
                    if value is not None:
                        data[n] = value
        for kind in errors:
            lnk = len(errors[kind])
            self.tm.error(
                '{} in lines {}'.format(
                    kind,
                    ','.join(str(ln) for ln in errors[kind][0:ERROR_CUTOFF])
                )
            )
            if lnk > ERROR_CUTOFF:
                self.tm.error(
                    '\t and {} more cases'.format(lnk - ERROR_CUTOFF),
                    tm=False
                )
        self.data = data
        if not errors:
            if self.fileName == WARP[0]:
                slotType = data[1]
                otype = []
                maxSlot = 1
                for n in sorted(data):
                    if data[n] == slotType:
                        maxSlot = n
                        continue
                    otype.append(data[n])
                otype.append(slotType)
                otype.append(maxSlot)
                self.data = tuple(otype)
            elif self.fileName == WARP[1]:
                slotsList = sorted(data)
                maxSlot = min(data.keys()) - 1
                oslots = []
                for n in slotsList:
                    oslots.append(tuple(sorted(data[n])))
                oslots.append(maxSlot)
                self.data = tuple(oslots)
        return not errors

    def _compute(self):
        good = True
        for feature in self.dependencies:
            if not feature.load():
                good = False
        if not good:
            return False

        def info(msg, tm=True):
            self.tm.info(cmpFormat.format(msg), tm=tm, cache=-1)

        cmpFormat = 'c {:<20} {{}}'.format(self.fileName)
        self.tm.indent(level=2, reset=True)

        def error(msg, tm=True):
            self.tm.error(cmpFormat.format(msg), tm=tm)

        self.data = self.method(
            info, error, *[
                dep.metaData if dep.fileName == WARP[2] else dep.data
                for dep in self.dependencies
            ]
        )
        good = self.data is not None
        if good:
            self.dataLoaded = time.time()
        return good

    def _writeTf(
        self,
        dirName=None,
        fileName=None,
        overwrite=True,
        extension=None,
        metaOnly=False,
        nodeRanges=False
    ):
        self.tm.indent(level=1, reset=True)
        metaOnly = metaOnly or self.isConfig

        dirName = dirName or self.dirName
        fileName = fileName or self.fileName
        extension = extension or self.extension
        if not os.path.exists(dirName):
            try:
                os.makedirs(dirName, exist_ok=True)
            except Exception:
                self.tm.error('Cannot create directory "{}"'.format(dirName))
                return False
        fpath = '{}/{}{}'.format(dirName, fileName, extension)
        if fpath == self.path:
            if os.path.exists(fpath):
                if not overwrite:
                    self.tm.error((
                        'Feature file "{}" already exists,'
                        ' feature will not be written'
                    ).format(fpath))
                    return False
        try:
            fh = open(fpath, 'w', encoding='utf8')
        except Exception:
            self.tm.error('Cannot write to feature file "{}"'.format(fpath))
            return False
        fh.write(
            '@{}\n'.format(
                'config' if self.isConfig else 'edge'
                if self.isEdge else 'node'
            )
        )
        if self.edgeValues:
            fh.write('@edgeValues\n')
        for meta in sorted(self.metaData):
            fh.write('@{}={}\n'.format(meta, self.metaData[meta]))
        fh.write('@writtenBy=Text-Fabric\n')
        fh.write(
            '@dateWritten={}\n'.
            format(datetime.utcnow().replace(microsecond=0).isoformat() + 'Z')
        )
        fh.write('\n')
        self._setDataType()
        good = True
        if not metaOnly:
            good = self._writeDataTf(fh, nodeRanges=nodeRanges)
        fh.close()
        msgFormat = '{:<1} {:<20} to {}'
        if good:
            self.tm.info(
                msgFormat.format('M' if metaOnly else 'T', fileName, dirName)
            )
        else:
            self.tm.error(
                msgFormat.format('M' if metaOnly else 'T', fileName, dirName)
            )
        return good

    def _writeDataTf(self, fh, nodeRanges=False):
        data = self.data
        if type(data) is tuple:
            maxSlot = data[-1]
            if self.fileName == WARP[0]:
                data = dict(((k + 1 + maxSlot, data[k])
                             for k in range(0,
                                            len(data) - 2)))
            elif self.fileName == WARP[1]:
                data = dict(((k + 1 + maxSlot, data[k])
                             for k in range(0,
                                            len(data) - 1)))
        edgeValues = self.edgeValues
        if self.isEdge:
            implicitNode = 1
            for n in sorted(data):
                thisData = data[n]
                sets = {}
                if edgeValues:
                    for m in thisData:
                        sets.setdefault(thisData[m], set()).add(m)
                    for (value, mset) in sorted(sets.items()):
                        nodeSpec2 = specFromRanges(rangesFromSet(mset))
                        nodeSpec = '' if n == implicitNode else n
                        implicitNode = n + 1
                        if value is None:
                            fh.write(
                                '{}{}{}\n'.format(
                                    nodeSpec,
                                    '\t' if nodeSpec else '',
                                    nodeSpec2,
                                )
                            )
                        else:
                            fh.write(
                                '{}{}{}\t{}\n'.format(
                                    nodeSpec,
                                    '\t' if nodeSpec else '',
                                    nodeSpec2,
                                    tfFromValue(value),
                                )
                            )
                else:
                    nodeSpec2 = specFromRanges(rangesFromSet(thisData))
                    nodeSpec = '' if n == implicitNode else n
                    implicitNode = n + 1
                    fh.write(
                        '{}{}{}\n'.format(
                            nodeSpec, '\t' if nodeSpec else '', nodeSpec2
                        )
                    )
        else:
            sets = {}
            if nodeRanges:
                for n in sorted(data):
                    sets.setdefault(data[n], []).append(n)
                implicitNode = 1
                for (value, nset) in sorted(
                    sets.items(), key=lambda x: (x[1][0], x[1][-1])
                ):
                    if len(nset) == 1 and nset[0] == implicitNode:
                        nodeSpec = ''
                    else:
                        nodeSpec = specFromRanges(rangesFromSet(nset))
                    implicitNode = nset[-1]
                    fh.write(
                        '{}{}{}\n'.format(
                            nodeSpec,
                            '\t' if nodeSpec else '',
                            tfFromValue(value),
                        )
                    )
            else:
                implicitNode = 1
                for n in sorted(data):
                    nodeSpec = '' if n == implicitNode else n
                    implicitNode = n + 1
                    fh.write(
                        '{}{}{}\n'.format(
                            nodeSpec,
                            '\t' if nodeSpec else '',
                            tfFromValue(data[n]),
                        )
                    )
        return True

    def _readDataBin(self):
        if not os.path.exists(self.binPath):
            self.tm.error(
                'TF reading: feature file "{}" does not exist'.format(
                    self.binPath
                )
            )
            return False
        with gzip.open(self.binPath, "rb") as f:
            self.data = pickle.load(f)
        self.dataLoaded = time.time()
        return True

    def cleanDataBin(self):
        if os.path.exists(self.binPath):
            os.unlink(self.binPath)

    def _writeDataBin(self):
        good = True
        if not os.path.exists(self.binDir):
            try:
                os.makedirs(self.binDir, exist_ok=True)
            except Exception:
                self.tm.error(
                    'Cannot create directory "{}"'.format(self.binDir)
                )
                good = False
        if not good:
            return False
        try:
            with gzip.open(self.binPath, "wb", compresslevel=GZIP_LEVEL) as f:
                pickle.dump(self.data, f, protocol=PICKLE_PROTOCOL)
        except Exception:
            self.tm.error('Cannot write to file "{}"'.format(self.binPath))
            good = False
        self.dataLoaded = time.time()
        return True

    def _getModified(self, bin=False):
        if bin:
            return os.path.getmtime(self.binPath) if os.path.exists(
                self.binPath
            ) else None
        else:
            if self.method:
                depsInfo = [dep._getModified() for dep in self.dependencies]
                depsModifieds = [d for d in depsInfo if d is not None]
                depsModified = None if len(depsModifieds
                                           ) == 0 else max(depsModifieds)
                if depsModified is not None:
                    return depsModified
                elif os.path.exists(self.binPath):
                    return os.path.getmtime(self.binPath)
                else:
                    return None
            else:
                if os.path.exists(self.path):
                    return os.path.getmtime(self.path)
                elif os.path.exists(self.binPath):
                    return os.path.getmtime(self.binPath)
                else:
                    return None



import os
import re
from IPython.display import display, Markdown, HTML

URL_GH = 'https://github.com'
URL_NB = 'http://nbviewer.jupyter.org/github'

CORPUS = 'BHSA'

SHEBANQ_URL = 'https://shebanq.ancient-data.org/hebrew'
SHEBANQ = (
    f'{SHEBANQ_URL}/text'
    '?book={book}&chapter={chapter}&verse={verse}&version={version}'
    '&mr=m&qw=q&tp=txt_p&tr=hb&wget=v&qget=v&nget=vt'
)
SHEBANQ_LEX = (f'{SHEBANQ_URL}/word' '?version={version}' '&id={lid}')

LIMIT_SHOW = 100
LIMIT_TABLE = 2000

CSS = '''
<style>
.verse {
    display: flex;
    flex-flow: row wrap;
    direction: rtl;
}
.vl {
    display: flex;
    flex-flow: column nowrap;
    direction: ltr;
}
.sentence,.clause,.phrase {
    margin-top: -1.2em;
    margin-left: 1em;
    background: #ffffff none repeat scroll 0 0;
    padding: 0 0.3em;
    border-style: solid;
    border-radius: 0.2em;
    font-size: small;
    display: block;
    width: fit-content;
    max-width: fit-content;
    direction: ltr;
}
.atoms {
    display: flex;
    flex-flow: row wrap;
    margin: 0.3em;
    padding: 0.3em;
    direction: rtl;
    background-color: #ffffff;
}
.satom,.catom,.patom {
    margin: 0.3em;
    padding: 0.3em;
    border-radius: 0.3em;
    border-style: solid;
    display: flex;
    flex-flow: column nowrap;
    direction: rtl;
    background-color: #ffffff;
}
.sentence {
    border-color: #aa3333;
    border-width: 1px;
}
.clause {
    border-color: #aaaa33;
    border-width: 1px;
}
.phrase {
    border-color: #33aaaa;
    border-width: 1px;
}
.satom {
    border-color: #aa3333;
    border-width: 4px;
}
.catom {
    border-color: #aaaa33;
    border-width: 3px;
}
.patom {
    border-color: #33aaaa;
    border-width: 3px;
}
.word {
    padding: 0.1em;
    margin: 0.1em;
    border-radius: 0.1em;
    border: 1px solid #cccccc;
    display: flex;
    flex-flow: column nowrap;
    direction: rtl;
    background-color: #ffffff;
}
.lex {
    padding: 0.1em;
    margin: 0.1em;
    border-radius: 0.1em;
    border: 2px solid #888888;
    width: fit-content;
    display: flex;
    flex-flow: column nowrap;
    direction: rtl;
    background-color: #ffffff;
}
.satom.l,.catom.l,.patom.l {
    border-left-style: dotted
}
.satom.r,.catom.r,.patom.r {
    border-right-style: dotted
}
.satom.L,.catom.L,.patom.L {
    border-left-style: none
}
.satom.R,.catom.R,.patom.R {
    border-right-style: none
}
.h,.h :visited,.h :link {
    font-family: "Ezra SIL", "SBL Hebrew", sans-serif;
    font-size: large;
    color: #000044;
    direction: rtl;
    text-decoration: none;
}
.rela,.function,.typ {
    font-family: monospace;
    font-size: small;
    color: #0000bb;
}
.sp,.sp :visited,.sp :link {
    font-family: monospace;
    font-size: medium;
    color: #0000bb;
    text-decoration: none;
}
.vl {
    font-family: monospace;
    font-size: medium;
    color: #0000bb;
}
.vvs {
    font-family: monospace;
    font-size: medium;
    font-weight: bold;
    color: #0000bb;
}
.vvt {
    font-family: monospace;
    font-size: medium;
    color: #0000bb;
}
.gl {
    font-family: sans-serif;
    font-size: small;
    color: #aaaaaa;
}
.vs {
    font-family: sans-serif;
    font-size: small;
    font-weight: bold;
    color: #444444;
}
.nd {
    font-family: monospace;
    font-size: x-small;
    color: #999999;
}
.hl {
    background-color: #ffee66;
}
</style>
'''

CLASS_NAMES = dict(
    verse='verse',
    sentence='atoms',
    sentence_atom='satom',
    clause='atoms',
    clause_atom='catom',
    phrase='atoms',
    phrase_atom='patom',
    subphrase='subphrase',
    word='word',
    lex='lex',
)

ATOMS = dict(
    sentence_atom='sentence',
    clause_atom='clause',
    phrase_atom='phrase',
)
SUPER = dict((y, x) for (x, y) in ATOMS.items())

SECTION = {'book', 'chapter', 'verse', 'half_verse'}


def _dm(md):
    display(Markdown(md))


def _outLink(text, href, title=None):
    titleAtt = '' if title is None else f' title="{title}"'
    return f'<a target="_blank" href="{href}"{titleAtt}>{text}</a>'


class Bhsa(object):
    def __init__(
        self,
        api,
        name,
        version='c',
    ):
        self.version = version
        api.TF.load(
            '''
            sp vs vt
            lex language gloss voc_lex voc_lex_utf8
            function typ rela
            number label book
        ''',
            add=True,
            silent=True
        )
        self.api = api
        self.cwd = os.getcwd()
        cwdPat = re.compile(f'^.*/github/([^/]+)/([^/]+)((?:/.+)?)$', re.I)
        cwdRel = cwdPat.findall(self.cwd)
        if cwdRel:
            (thisOrg, thisRepo, thisPath) = cwdRel[0]
            onlineTail = (
                f'{thisOrg}/{thisRepo}'
                f'/blob/master{thisPath}/{name}.ipynb'
            )
        else:
            cwdRel = None
        nbLink = (
            None
            if name is None or cwdRel is None else f'{URL_NB}/{onlineTail}'
        )
        ghLink = (
            None
            if name is None or cwdRel is None else f'{URL_GH}/{onlineTail}'
        )
        docLink = f'https://etcbc.github.io/bhsa'
        extraLink = f'https://github.com/Dans-labs/text-fabric/wiki/Bhsa'
        dataLink = _outLink(CORPUS, docLink, '{provenance of this corpus}')
        featureLink = _outLink(
            'Feature docs',
            f'{docLink}/features/hebrew/{self.version}/0_home.html',
            '{CORPUS} feature documentation'
        )
        bhsaLink = _outLink('BHSA API', extraLink, 'BHSA API documentation')
        tfLink = _outLink(
            f'Text-Fabric API {api.TF.version}',
            'https://github.com/Dans-labs/text-fabric/wiki/api',
            'text-fabric-api'
        )
        tfsLink = _outLink(
            'Search Reference',
            (
                'https://github.com/Dans-labs/text-fabric/wiki/api'
                '#search-template-introduction'
            ),
            'Search Templates Introduction and Reference'
        )
        _dm(
            '**Documentation:**'
            f' {dataLink} {featureLink} {bhsaLink} {tfLink} {tfsLink}'
        )
        if nbLink:
            _dm(
                f'''
This notebook online:
{_outLink('NBViewer', nbLink)}
{_outLink('GitHub', ghLink)}
'''
            )
        self._loadCSS()

    def shbLink(self, n, text=None, asString=False):
        api = self.api
        L = api.L
        T = api.T
        F = api.F
        version = self.version
        nType = F.otype.v(n)
        if nType == 'lex':
            lex = F.lex.v(n)
            lan = F.language.v(n)
            lexId = '{}{}'.format(
                '1' if lan == 'Hebrew' else '2',
                lex.replace('>', 'A').replace('<',
                                              'O').replace('[', 'v').replace(
                                                  '/', 'n'
                                              ).replace('=', 'i'),
            )
            href = SHEBANQ_LEX.format(
                version=version,
                lid=lexId,
            )
            title = 'show this lexeme in SHEBANQ'
            if text is None:
                text = F.voc_lex_utf8.v(n)
            result = _outLink(text, href, title=title)
            if asString:
                return result
            display(HTML(result))
            return

        (bookE, chapter, verse) = T.sectionFromNode(n)
        bookNode = n if nType == 'book' else L.u(n, otype='book')[0]
        book = F.book.v(bookNode)
        passageText = (
            bookE if nType == 'book' else '{} {}'.format(bookE, chapter) if
            nType == 'chapter' else '{} {}:{}'.format(bookE, chapter, verse)
        )
        href = SHEBANQ.format(
            version=version,
            book=book,
            chapter=chapter,
            verse=verse,
        )
        if text is None:
            text = passageText
            title = 'show this passage in SHEBANQ'
        else:
            title = passageText
        result = _outLink(text, href, title=title)
        if asString:
            return result
        display(HTML(result))

    def plain(
        self,
        n,
        linked=True,
        withNodes=False,
        asString=False,
    ):
        api = self.api
        L = api.L
        T = api.T
        F = api.F

        nType = F.otype.v(n)
        markdown = ''
        nodeRep = f' *{n}* ' if withNodes else ''

        if nType == 'word':
            rep = T.text([n])
            if linked:
                rep = self.shbLink(n, text=rep, asString=True)
        elif nType in SECTION:
            fmt = (
                '{}' if nType == 'book'
                else '{} {}' if nType == 'chapter'
                else '{} {}:{}'
            )
            rep = fmt.format(*T.sectionFromNode(n))
            if nType == 'half_verse':
                rep += F.label.v(n)
            if linked:
                rep = self.shbLink(n, text=rep, asString=True)
        elif nType == 'lex':
            rep = F.voc_lex_utf8.v(n)
            if linked:
                rep = self.shbLink(n, text=rep, asString=True)
        else:
            rep = T.text(L.d(n, otype='word'))
            if linked:
                rep = self.shbLink(n, text=rep, asString=True)

        markdown = f'{rep}{nodeRep}'

        if asString:
            return markdown
        _dm((markdown))

    def plainTuple(
        self,
        ns,
        seqNumber,
        linked=1,
        withNodes=False,
        asString=False,
    ):
        markdown = [str(seqNumber)]
        for (i, n) in enumerate(ns):
            markdown.append(
                self.plain(
                    n,
                    linked=i == linked - 1,
                    withNodes=withNodes,
                    asString=True,
                )
            )
        if asString:
            return ' | '.join(markdown)
        _dm(' , '.join(markdown))

    def pretty(self, n, withNodes=False, suppress=set(), highlights=set()):
        html = []
        self._pretty(
            n,
            html,
            firstSlot=None,
            lastSlot=None,
            withNodes=withNodes,
            suppress=suppress,
            highlights=highlights,
        )
        htmlStr = '\n'.join(html)
        display(HTML(htmlStr))

    def prettyTuple(
        self,
        ns,
        seqNumber,
        item='Result',
        withNodes=False,
        suppress=set(),
    ):
        api = self.api
        L = api.L
        F = api.F
        sortNodes = api.sortNodes
        verses = set()
        lexemes = set()
        highlights = set()
        for n in ns:
            nType = F.otype.v(n)
            if nType not in SECTION:
                firstSlot = n if nType == 'word' else L.d(n, otype='word')[0]
                if nType == 'lex':
                    lexemes.add(n)
                    highlights.add(n)
                else:
                    verses.add(L.u(firstSlot, otype='verse')[0])
                    atomType = SUPER.get(nType, None)
                    if atomType:
                        highlights |= set(L.d(n, otype=atomType))
                    else:
                        highlights.add(n)
        _dm(f'''
##### {item} {seqNumber}
''')
        for l in sortNodes(lexemes):
            self.pretty(
                l,
                withNodes=withNodes,
                suppress=suppress,
                highlights=highlights,
            )
        for v in sortNodes(verses):
            self.pretty(
                v,
                withNodes=withNodes,
                suppress=suppress,
                highlights=highlights,
            )

    def search(self, query):
        api = self.api
        S = api.S
        results = list(S.search(query))
        nResults = len(results)
        plural = '' if nResults == 1 else 's'
        print(f'{nResults} result{plural}')
        return results

    def table(
        self,
        results,
        start=None,
        end=None,
        linked=1,
        withNodes=False,
        asString=False,
    ):
        api = self.api
        F = api.F

        collected = []
        if start is None:
            start = 1
        i = -1
        rest = 0
        if not hasattr(results, 'len'):
            if end is None or end > LIMIT_TABLE:
                end = LIMIT_TABLE
            for result in results:
                i += 1
                if i < start - 1:
                    continue
                if i >= end:
                    break
                collected.append((i + 1, result))
        else:
            if end is None:
                end = len(results)
            rest = 0
            if end - (start - 1) > LIMIT_TABLE:
                rest = end - (start - 1) - LIMIT_TABLE
                end = start - 1 + LIMIT_TABLE
            for i in range(start - 1, end):
                collected.append((i + 1, results[i]))

        if len(collected) == 0:
            return
        (firstSeq, firstResult) = collected[0]
        nColumns = len(firstResult)
        markdown = [
            'n | '
            +
            (' | '.join(F.otype.v(n) for n in firstResult))
        ]
        markdown.append(' | '.join('---' for n in range(nColumns + 1)))
        for (seqNumber, ns) in collected:
            markdown.append(
                self.plainTuple(
                    ns,
                    seqNumber,
                    linked=linked,
                    withNodes=withNodes,
                    asString=True,
                )
            )
        markdown = '\n'.join(markdown)
        if asString:
            return markdown
        _dm(markdown)
        if rest:
            _dm(
                f'**{rest} more results skipped**'
                f' because we show a maximum of'
                f' {LIMIT_TABLE} results at a time'
            )

    def show(
        self,
        results,
        condensed=True,
        start=None,
        end=None,
        withNodes=False,
        suppress=set()
    ):
        if condensed:
            (passages, verses) = self._condense(results)
            results = passages if len(verses) == 0 else verses
        if start is None:
            start = 1
        i = -1
        if not hasattr(results, 'len'):
            if end is None or end > LIMIT_SHOW:
                end = LIMIT_SHOW
            for result in results:
                i += 1
                if i < start - 1:
                    continue
                if i >= end:
                    break
                self.prettyTuple(
                    result,
                    i + 1,
                    item='Verse' if condensed else 'Result',
                    withNodes=withNodes,
                    suppress=suppress
                )
        else:
            if end is None:
                end = len(results)
            rest = 0
            if end - (start - 1) > LIMIT_SHOW:
                rest = end - (start - 1) - LIMIT_SHOW
                end = start - 1 + LIMIT_SHOW
            for i in range(start - 1, end):
                self.prettyTuple(
                    results[i],
                    i + 1,
                    item='Passage' if condensed else 'Result',
                    withNodes=withNodes,
                    suppress=suppress
                )
            if rest:
                _dm(
                    f'**{rest} more results skipped**'
                    f' because we show a maximum of'
                    f' {LIMIT_SHOW} results at a time'
                )

    def _condense(self, results):
        api = self.api
        F = api.F
        L = api.L
        sortNodes = api.sortNodes
        verses = {}
        passages = set()
        for ns in results:
            for n in ns:
                nType = F.otype.v(n)
                if nType in SECTION:
                    if nType == 'verse':
                        verses.setdefault(n, set())
                    passages.add(n)
                else:
                    fw = n if nType == 'word' else L.d(n, otype='word')[0]
                    v = L.u(fw, otype='verse')[0]
                    verses.setdefault(v, set()).add(n)
        return (
            tuple(sortNodes(passages)),
            tuple((p,) + tuple(verses[p]) for p in sortNodes(verses)),
        )

    def _pretty(
        self,
        n,
        html,
        firstSlot=None,
        lastSlot=None,
        withNodes=True,
        suppress=set(),
        highlights=set()
    ):
        api = self.api
        L = api.L
        T = api.T
        F = api.F
        sortNodes = api.sortNodes
        hl = ' hl' if n in highlights else ''
        nType = F.otype.v(n)
        className = CLASS_NAMES.get(nType, None)
        superType = ATOMS.get(nType, None)
        boundaryClass = ''
        (myStart, myEnd) = (n, n) if nType == 'word' else self._getBoundary(n)
        if superType:
            (superNode, superStart, superEnd) = self._getSuper(n, superType)
            if superStart < myStart:
                boundaryClass += ' r'
            if superEnd > myEnd:
                boundaryClass += ' l'
        if ((firstSlot and myEnd < firstSlot)
            or (lastSlot and myStart > lastSlot)):  # noqa 129
            return
        if firstSlot and (myStart < firstSlot):
            boundaryClass += ' R'
        if lastSlot and (myEnd > lastSlot):
            boundaryClass += ' L'
        if nType == 'book':
            html.append(self.shbLink(n, asString=True))
            return
        elif nType == 'chapter':
            html.append(self.shbLink(n, asString=True))
            return
        elif nType == 'half_verse':
            html.append(self.shbLink(n, asString=True))
            return
        elif nType == 'verse':
            label = self.shbLink(n, asString=True)
            (firstSlot, lastSlot) = self._getBoundary(n)
            children = sortNodes(
                set(L.d(n, otype='sentence_atom')) | {
                    L.u(firstSlot, otype='sentence_atom')[0],
                    L.u(lastSlot, otype='sentence_atom')[0],
                }
            )
        elif nType == 'sentence':
            children = L.d(n, otype='sentence_atom')
        elif nType == 'sentence_atom' or nType == 'clause':
            children = L.d(n, otype='clause_atom')
        elif nType == 'clause_atom' or nType == 'phrase':
            children = L.d(n, otype='phrase_atom')
        elif nType == 'phrase_atom' or nType == 'subphrase':
            children = L.d(n, otype='word')
        elif nType == 'lex':
            children = ()
        elif nType == 'word':
            children = ()
            lx = L.u(n, otype='lex')[0]
        html.append(f'''<div class="{className} {boundaryClass} {hl}">''')
        if nType == 'verse':
            nodePart = f'<div class="nd">{n}</div>' if withNodes else ''
            html.append(
                f'''
    <div class="vl">
        <div class="vs">{label}</div>{nodePart}
    </div>
'''
            )
        elif nType == 'word':
            lx = L.u(n, otype='lex')[0]
            lexLink = (
                self.shbLink(lx, text=T.text([n]), asString=True)
            )
            html.append(f'<div class="h">{lexLink}</div>')
            if 'sp' not in suppress:
                spLink = (
                    self.shbLink(n, text=F.sp.v(n), asString=True)
                )
                html.append(
                    f'<div class="sp">{spLink}</div>'
                )
            if 'gloss' not in suppress:
                html.append(
                    f'<div class="gl">'
                    f'{F.gloss.v(lx).replace("<", "&lt;")}</div>'
                )
            if F.sp.v(n) == 'verb':
                if 'vs' not in suppress:
                    html.append(f'<div class="vvs">' f'{F.vs.v(n)}</div>')
                if 'vt' not in suppress:
                    html.append(f'<div class="vvt">' f'{F.vt.v(n)}</div>')
        elif nType == 'lex':
            html.append(f'<div class="h">{F.voc_lex_utf8.v(n)}</div>')
            if 'voc_lex' not in suppress:
                llink = self.shbLink(
                    n, text=F.voc_lex.v(n).replace("<", "&lt;"),
                    asString=True
                )
                html.append(f'<div class="vl">{llink}</div>')
            if 'gloss' not in suppress:
                html.append(
                    f'<div class="gl">'
                    f'{F.gloss.v(n).replace("<", "&lt;")}</div>'
                )
        elif superType:
            nodePart = (
                f'<span class="nd">{superNode}</span>' if withNodes else ''
            )
            typePart = self.shbLink(superNode, text=superType, asString=True)
            featurePart = ''
            if superType == 'clause':
                if 'rela' not in suppress:
                    featurePart += (
                        f' <span class="rela">{F.rela.v(superNode)}</span>'
                    )
                if 'typ' not in suppress:
                    featurePart += (
                        f' <span class="typ">{F.typ.v(superNode)}</span>'
                    )
            elif superType == 'phrase':
                if 'function' not in suppress:
                    featurePart += (
                        f' <span class="function">'
                        f'{F.function.v(superNode)}</span>'
                    )
                if 'typ' not in suppress:
                    featurePart += (
                        f' <span class="typ">{F.typ.v(superNode)}</span>'
                    )
            html.append(
                f'''
    <div class="{superType}">{typePart} {nodePart} {featurePart}
    </div>
    <div class="atoms">
'''
            )
        for ch in children:
            self._pretty(
                ch,
                html,
                firstSlot=firstSlot,
                lastSlot=lastSlot,
                withNodes=withNodes,
                suppress=suppress,
                highlights=highlights,
            )
        if superType:
            html.append('''
    </div>
''')
        html.append('''
</div>
''')

    def _getBoundary(self, n):
        api = self.api
        L = api.L
        words = L.d(n, otype='word')
        return (words[0], words[-1])

    def _getSuper(self, n, tp):
        api = self.api
        L = api.L
        superNode = L.u(n, otype=tp)[0]
        return (superNode, *self._getBoundary(superNode))

    def _loadCSS(self):
        display(HTML(CSS))



import collections
from .helpers import makeInverse, makeInverseVal
from .locality import Locality
from .text import Text
from .search import Search


class OtypeFeature(object):
    def __init__(self, api, data=None):
        self.api = api
        self.data = data
        self.slotType = self.data[-2]
        self.maxSlot = self.data[-1]
        self.maxNode = len(self.data) - 2 + self.maxSlot

    def v(self, n):
        if n == 0:
            return None
        if n < self.maxSlot + 1:
            return self.data[-2]
        m = n - self.maxSlot
        if m <= len(self.data) - 2:
            return self.data[m - 1]
        return None

    def s(self, val):
        # NB: the support attribute has been added by precomputing __levels__
        if val in self.support:
            (b, e) = self.support[val]
            return range(b, e + 1)
        else:
            return ()

    def sInterval(self, val):
        # NB: the support attribute has been added by precomputing __levels__
        if val in self.support:
            return self.support[val]
        else:
            return ()


class OslotsFeature(object):
    def __init__(self, api, data=None):
        self.api = api
        self.data = data
        self.maxSlot = self.data[-1]

    def s(self, n):
        if n == 0:
            return ()
        if n < self.maxSlot + 1:
            return (n, )
        m = n - self.maxSlot
        if m <= len(self.data) - 1:
            return self.data[m - 1]
        return ()


class NodeFeature(object):
    def __init__(self, api, data):
        self.api = api
        self.data = data

    def v(self, n):
        if n in self.data:
            return self.data[n]
        return None

    def s(self, val):
        Crank = self.api.C.rank.data
        return tuple(
            sorted(
                [n for n in self.data if self.data[n] == val],
                key=lambda n: Crank[n - 1],
            )
        )

    def freqList(self):
        fql = collections.Counter()
        for n in self.data:
            fql[self.data[n]] += 1
        return tuple(sorted(fql.items(), key=lambda x: (-x[1], x[0])))


class EdgeFeature(object):
    def __init__(self, api, data, doValues):
        self.api = api
        self.data = data
        self.doValues = doValues
        self.dataInv = makeInverseVal(self.data) if doValues else makeInverse(
            self.data
        )

    def f(self, n):
        Crank = self.api.C.rank.data
        if n in self.data:
            if self.doValues:
                return tuple(
                    sorted(
                        self.data[n].items(),
                        key=lambda mv: Crank[mv[0] - 1],
                    )
                )
            else:
                return tuple(
                    sorted(
                        self.data[n],
                        key=lambda m: Crank[m - 1],
                    )
                )
        return ()

    def t(self, n):
        Crank = self.api.C.rank.data
        if n in self.dataInv:
            if self.doValues:
                return tuple(
                    sorted(
                        self.dataInv[n].items(),
                        key=lambda mv: Crank[mv[0] - 1],
                    )
                )
            else:
                return tuple(
                    sorted(
                        self.dataInv[n],
                        key=lambda m: Crank[m - 1],
                    )
                )
        return ()


class Computed(object):
    def __init__(self, api, data):
        self.api = api
        self.data = data


class NodeFeatures(object):
    pass


class EdgeFeatures(object):
    pass


class Computeds(object):
    pass


class Api(object):
    def __init__(self, TF):
        self.TF = TF
        self.ignored = tuple(sorted(TF.featuresIgnored))
        self.F = NodeFeatures()
        self.Feature = self.F
        self.E = EdgeFeatures()
        self.Edge = self.E
        self.C = Computeds()
        self.Computed = self.C
        self.info = TF.tm.info
        self.error = TF.tm.error
        self.indent = TF.tm.indent
        self.loadLog = TF.tm.cache
        setattr(self, 'FeatureString', self.Fs)
        setattr(self, 'EdgeString', self.Es)
        setattr(self, 'ComputedString', self.Cs)
        setattr(self, 'Nodes', self.N)
        setattr(self, 'AllFeatures', self.Fall)
        setattr(self, 'AllEdges', self.Eall)
        setattr(self, 'AllComputeds', self.Call)

    def Fs(self, fName):
        if not hasattr(self.F, fName):
            self.error('Node feature "{}" not loaded'.format(fName))
            return None
        return getattr(self.F, fName)

    def Es(self, fName):
        if not hasattr(self.E, fName):
            self.error('Edge feature "{}" not loaded'.format(fName))
            return None
        return getattr(self.E, fName)

    def Cs(self, fName):
        if not hasattr(self.C, fName):
            self.error('Computed feature "{}" not loaded'.format(fName))
            return None
        return getattr(self.C, fName)

    def N(self):
        for n in self.C.order.data:
            yield n

    def sortNodes(self, nodeSet):
        Crank = self.C.rank.data
        return sorted(nodeSet, key=lambda n: Crank[n - 1])

    def Fall(self):
        return sorted(x[0] for x in self.F.__dict__.items())

    def Eall(self):
        return sorted(x[0] for x in self.E.__dict__.items())

    def Call(self):
        return sorted(x[0] for x in self.C.__dict__.items())

    def makeAvailableIn(self, scope):
        for member in dir(self):
            if '_' not in member and member != 'makeAvailableIn':
                scope[member] = getattr(self, member)


def addSortKey(api):
    Crank = api.C.rank.data
    api.sortKey = lambda n: Crank[n - 1]


def addOtype(api):
    setattr(api.F.otype, 'all', tuple(o[0] for o in api.C.levels.data))
    setattr(
        api.F.otype, 'support',
        dict(((o[0], (o[2], o[3])) for o in api.C.levels.data))
    )


def addLocality(api):
    api.L = Locality(api)
    api.Locality = api.L


def addText(api):
    api.T = Text(api)
    api.Text = api.T


def addSearch(api, silent):
    api.S = Search(api, silent)
    api.Search = api.S



from .data import WARP
from .helpers import itemize, compileFormats

DEFAULT_FORMAT = 'text-orig-full'


class Text(object):
    def __init__(self, api):
        self.api = api
        self.languages = {}
        self.nameFromNode = {}
        self.nodeFromName = {}
        config = api.TF.features[WARP[2]].metaData if WARP[
            2
        ] in api.TF.features else {}
        self.sectionTypes = itemize(config.get('sectionTypes', ''), ',')
        sectionFeats = itemize(config.get('sectionFeatures', ''), ',')
        self.sectionFeatures = []
        self.config = config

        good = True
        if len(sectionFeats) != 0 and len(self.sectionTypes) != 0:
            for (fName, fObj) in sorted(
                f for f in api.TF.features.items()
                if f[0] == sectionFeats[0]
                or f[0].startswith('{}@'.format(sectionFeats[0]))
            ):
                if not fObj.load(silent=True):
                    good = False
                    continue
                meta = fObj.metaData
                code = meta.get('languageCode', '')
                self.languages[code] = dict(
                    ((k, meta.get(k, 'default'))
                     for k in ('language', 'languageEnglish'))
                )
                self.nameFromNode[code] = fObj.data
                self.nodeFromName[code] = dict(
                    ((name, node) for (node, name) in fObj.data.items())
                )
            for fName in (sectionFeats):
                if not api.TF.features[fName].load(silent=True):
                    good = False
                    continue
                self.sectionFeatures.append(api.TF.features[fName].data)

            sec0 = self.sectionTypes[0]
            setattr(self, '{}Name'.format(sec0), self._sec0Name)
            setattr(self, '{}Node'.format(sec0), self._sec0Node)

        self._xformats = compileFormats(api.TF._cformats, api.TF.features)
        self.formats = set(self._xformats.keys())
        self.good = good

    def _sec0Name(self, n, lang='en'):
        sec0T = self.sectionTypes[0]
        return self.nameFromNode[''
                                 if lang not in self.languages else lang].get(
                                     n if self.api.F.otype.v(n) == sec0T else
                                     self.api.L.u(n, sec0T)[0],
                                     'not a {} node'.format(sec0T),
                                 )

    def _sec0Node(self, name, lang='en'):
        return self.nodeFromName['' if lang not in self.languages else lang
                                 ].get(name, None)

    def sectionFromNode(self, n, lastSlot=False, lang='en'):
        sTypes = self.sectionTypes
        if len(sTypes) == 0:
            return ()
        sFs = self.sectionFeatures
        F = self.api.F
        L = self.api.L
        slotType = F.otype.slotType
        nType = F.otype.v(n)
        r = L.d(n, slotType)[-1] if lastSlot and nType != slotType else L.d(
            n, slotType
        )[0] if nType != slotType else n
        n0s = L.u(r, sTypes[0])
        n0 = n0s[0] if n0s else None
        n1s = L.u(r, sTypes[1])
        n1 = n1s[0] if n1s else None
        n2s = L.u(r, sTypes[2])
        n2 = n2s[0] if n2s else None
        return (
            self._sec0Name(n0, lang=lang),
            sFs[1].get(n1, None),
            sFs[2].get(n2, None),
        )

    def nodeFromSection(self, section, lang='en'):
        sTypes = self.sectionTypes
        if len(sTypes) == 0:
            return None
        (sec1, sec2) = self.api.C.sections.data
        sec0node = self._sec0Node(section[0], lang=lang)
        if len(section) == 1:
            return sec0node
        elif len(section) == 2:
            return sec1.get(sec0node, {}).get(section[1], None)
        else:
            return sec2.get(sec0node, {}).get(section[1],
                                              {}).get(section[2], None)

    def text(self, slots, fmt=None):
        if fmt is None:
            fmt = DEFAULT_FORMAT
        repf = self._xformats.get(fmt, None)
        if repf is None:
            return ' '.join(str(s) for s in slots)
        return ''.join(repf(s) for s in slots)



from distutils.core import setup

setup(
    name='text-fabric',
    packages=['tf', 'tf.extra'],
    version='3.4.12',
    description='''Processor for Text Fabric Data''',
    author='Dirk Roorda',
    author_email='dirk.roorda@dans.knaw.nl',
    url='https://github.com/Dans-labs/text-fabric',
    keywords=[
        'text', 'linguistics', 'database', 'graph', 'hebrew', 'greek', 'syriac'
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Education",
        "Intended Audience :: Religion",
        "Intended Audience :: Science/Research",
        "License :: Free for non-commercial use",
        "Natural Language :: English",
        "Natural Language :: Hebrew",
        "Natural Language :: Greek",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Religion",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Text Processing :: Markup :: XML",
    ],
    long_description='''\
Tools to read Text-Fabric resources analyse them efficiently.
More info on https://github.com/Dans-labs/text-fabric/wiki
''',
)
